###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import sys
import logging

import dendropy

from biolib.newick import parse_label, create_label


class RerootTree(object):
    """Reroot tree."""

    def __init__(self):
        """Initialization."""
        self.logger = logging.getLogger()
        
    def _reroot(self, tree, outgroup_node, max_support=100):
        """Reroot tree taking proper care of bootstrap values."""

        # determine support values for each bipartition
        tree.encode_bipartitions()
        support_values = {}
        for nd in tree:
            support, taxon, aux_info = parse_label(nd.label)
            if nd.is_leaf():
                support_values[nd.bipartition] = max_support
            else:
                if support is not None:
                    support_values[nd.bipartition] = float(support)
                else:
                    support_values[nd.bipartition] = None
     
        # move support values for desired re-rooting
        new_root = outgroup_node.parent_node
        tree.reseed_at(new_root)
        tree.encode_bipartitions()
        for nd in tree:
            _, taxon, aux_info = parse_label(nd.label)
            nd.label = create_label(support_values.get(nd.bipartition, "not_specified"), taxon, aux_info)
        tree.seed_node.edge.length = None
     
        # do a hard re-rooting of the tree
        # (this invalidates the previous bipartitions, so must be done seperately)
        tree.is_rooted = True
        tree.reroot_at_edge(outgroup_node.edge,
                                    length1=0.5 * outgroup_node.edge_length,
                                    length2=0.5 * outgroup_node.edge_length)

        # determine bootstrap for new node
        for child in tree.seed_node.child_node_iter():
            if outgroup_node.is_leaf():
                if not child.is_leaf():
                    support, taxon, aux_info = parse_label(child.label)
                    child.label = create_label(max_support, taxon, aux_info)
            else:
                if child != outgroup_node:
                    support, _taxon, _aux_info = parse_label(outgroup_node.label)
                    _support, taxon, aux_info = parse_label(child.label)
                    child.label = create_label(support, taxon, aux_info)
 
        return tree

    def root_with_outgroup(self, input_tree, output_tree, outgroup):
        """Reroot the tree using the given outgroup.

        Parameters
        ----------
        input_tree : str
          File containing Newick tree to rerooted.
        output_tree : str
          Name of file for rerooted tree.
        outgroup : iterable
          Labels of taxa in outgroup.
        """

        outgroup_in_tree = set()
        tree = dendropy.Tree.get_from_path(input_tree, 
                                            schema='newick', 
                                            rooting='force-unrooted', 
                                            preserve_underscores=True)

        outgroup = set(outgroup)
        for n in tree.leaf_node_iter():
            if n.taxon.label in outgroup:
                outgroup_in_tree.add(n.taxon)

        self.logger.info('Identified %d outgroup taxa in the tree.' % len(outgroup_in_tree))

        if len(outgroup_in_tree) == 0:
            self.logger.error('No outgroup taxa identified in the tree.')
            self.logger.error('Tree was not rerooted.')
            sys.exit(-1)

        mrca = tree.mrca(taxa=outgroup_in_tree)

        if len(mrca.leaf_nodes()) != len(outgroup_in_tree):
            self.logger.info('Outgroup is not monophyletic. Tree will be rerooted at the MRCA of the outgroup.')
            self.logger.info('The outgroup consisted of %d taxa, while the MRCA has %d leaf nodes.' % (len(outgroup_in_tree), len(mrca.leaf_nodes())))
            if len(mrca.leaf_nodes()) == len(tree.leaf_nodes()):
                self.logger.warning('The MRCA spans all taxa in the tree.')
                self.logger.warning('This indicating the selected outgroup is likely polyphyletic in the current tree.')
                self.logger.warning('Polyphyletic outgroups are not suitable for rooting. Try another outgroup.')
        else:
            self.logger.info('Outgroup is monophyletic.')

        if mrca.edge_length is None:
            self.logger.info('Tree appears to already be rooted on this outgroup.')
        else:
            self.logger.info('Rerooting tree.')
            self._reroot(tree, mrca)
            tree.write_to_path(output_tree, 
                                schema='newick', 
                                suppress_rooting=True, 
                                unquoted_underscores=True)
            self.logger.info('Rerooted tree written to: %s' % output_tree)

    def midpoint(self, input_tree, output_tree):
        """Reroot tree bat midpoint.

        Parameters
        ----------
        input_tree : str
          File containing Newick tree to rerooted.
        output_tree : str
          Name of file for rerooted tree.
        """

        tree = dendropy.Tree.get_from_path(input_tree, 
                                            schema='newick', 
                                            rooting='force-rooted', 
                                            preserve_underscores=True)
        tree.reroot_at_midpoint()
        tree.write_to_path(output_tree, 
                            schema='newick', 
                            suppress_rooting=True, 
                            unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import shutil
import logging
import ntpath
from collections import defaultdict

from genometreetk.default_values import DefaultValues
from genometreetk.markers.align_markers import AlignMarkers
from genometreetk.common import read_genome_id_file, read_genome_dir_file

from biolib.external.fasttree import FastTree

import dendropy

import pickle  # ***


class InferMarkers(object):
    """Identify ubiquitous, single-copy marker gene within a set of genomes.

    Currently, this class is specifically designed to work with reference
    genomes from NCBI. In particular, it assumes each genome is contain in
    a separate directory and that Pfam and TIGRFAMs annotations are
    available in a specific format.
    """

    def __init__(self, genome_dir_file, pfam_model_file, tigrfams_model_dir, cpus):
        """Initialization.

        Parameters
        ----------
        genome_dir_file : str
            File specifying directory for each genome.
        pfam_model_file : str
            File containing Pfam HMMs.
        tigrfams_model_dir : str
            Directory containing TIGRFAMs HMMs.
        cpus : int
            Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.genome_dir_file = genome_dir_file
        self.pfam_model_file = pfam_model_file
        self.tigrfams_model_dir = tigrfams_model_dir

        self.cpus = cpus

        self.protein_file_ext = DefaultValues.PROTEIN_FILE_EXTENSION
        self.pfam_extension = DefaultValues.PFAM_EXTENSION
        self.tigr_extension = DefaultValues.TIGR_EXTENSION

    def _read_gene_hits(self, table, genome_ids, genome_dirs, extension):
        """Read gene annotations from top hit file.

        Parameters
        ----------
        table : defaultdict(lambda: defaultdict(set))
            Table to populate with gene information.
        genome_ids : iterable
            Genomes of interest.
        genome_dirs : d[assembly_accession] -> directory
            Path to files for individual genomes.
        extension : str
            Extension of file containing top hits to each gene.
        """

        for genome_id in genome_ids:
            genome_dir = genome_dirs[genome_id]

            assembly = genome_dir[genome_dir.rfind('/') + 1:]
            tophit_file = os.path.join(genome_dir, assembly + extension)

            with open(tophit_file) as f:
                f.readline()

                for line in f:
                    line_split = line.split('\t')

                    gene_id = line_split[0]
                    hits = line_split[1].split(';')
                    for hit in hits:
                        protein_family, _evalue, _bitscore = hit.split(',')
                        table[protein_family][genome_id].add(gene_id)

    def _gene_count_table(self, genome_ids, genome_dirs):
        """Get Pfam and TIGRFAMs annotations for genomes.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        genome_dirs : d[assembly_accession] -> directory
            Path to files for individual genomes.

        Returns
        -------
        d[assembly_accession][genome_id] -> set([gene_id_1, ..., gene_id_N])
            Gene location of protein families within each genome.
        """

        table = defaultdict(lambda: defaultdict(set))
        self._read_gene_hits(table, genome_ids, genome_dirs, self.pfam_extension)
        self._read_gene_hits(table, genome_ids, genome_dirs, self.tigr_extension)

        return table

    def _marker_genes(self, genome_ids, gene_count_table, ubiquity_threshold, single_copy_threshold, output_file):
        """Identify genes meeting ubiquity and single-copy thresholds.

        Parameters
        ----------
        genome_ids : iterable
            Genomes of interest.
        gene_count_table : d[family_id][genome_id] -> set([gene_id_1, ..., gene_id_N])
            Gene location of protein families within each genome.
        ubiquity_threshold : float
            Threshold for defining a ubiquitous marker genes [0, 1].
        single_copy_threshold : float
            Threshold for defining a single-copy marker gene [0, 1].
        output_file : str
            Output file indicating ubiquity and single-copy values for all genes.

        Returns
        -------
        dict: d[protein family] -> (ubiquity, single copy)
            Marker genes satisfying selection criteria.
        """

        if ubiquity_threshold > 1 or single_copy_threshold > 1:
            self.logger.error('Ubiquity or single-copy threshold is invalid: %f, %f' % (ubiquity_threshold, single_copy_threshold))
            sys.exit(0)

        fout = open(output_file, 'w')
        fout.write('Model accession\tUbiquity\tSingle copy\n')

        # find genes meeting ubiquity and single-copy thresholds
        markers = {}
        for protein_family, genes_in_genomes in gene_count_table.items():
            ubiquity = 0
            single_copy = 0

            for genome_id in genome_ids:
                count = len(genes_in_genomes.get(genome_id, []))

                if count > 0:
                    ubiquity += 1

                if count == 1:
                    single_copy += 1

            u = ubiquity * 100.0 / len(genome_ids)
            s = single_copy * 100.0 / ubiquity
            fout.write('%s\t%.1f\t%.1f\n' % (protein_family, u, s))

            if ubiquity >= (ubiquity_threshold * len(genome_ids)) and single_copy >= (single_copy_threshold * ubiquity):
                markers[protein_family] = (u, s)

        fout.close()

        return markers

    def _identify_redundant_hmms(self, marker_genes, gene_count_table, redundancy, output_file):
        """Identify HMMs that consistently hit the same gene.

        This function identifies redundant HMMs both between and within
        the set of Pfam and TIGRFAMs HMMs. Redundancy between these two
        protein families is common. Redundancy within the TIGRFAMs HMMs
        is also common as there are a number of lineage-specific HMMs
        (e.g., archaeal and universal). Preference is given to TIGRFAM
        HMMs over Pfam HMMs as the former often models full genes instead
        of just domains.

        Parameters
        ----------
        marker_genes : iterable
            Marker genes to process for redundancy.
        gene_count_table : d[family_id][genome_id] -> set([gene_id_1, ..., gene_id_N])
            Gene location of protein families within each genome.
        redundancy : float
            Threshold for declaring HMMs redundant.
        output_file : str
            Output file to contain list of HMMs deemed to be redundant.

        Returns
        -------
        set
            Marker genes identified as being redundant.
        """

        if redundancy < 1:
            self.logger.error('Redundancy threshold is invalid: %f.' % redundancy)
            sys.exit(0)

        fout = open(output_file, 'w')
        fout.write('Kept marker\tRedundant marker\n')

        marker_gene_list = list(marker_genes)

        # count number of times HMMs hit the same gene
        redundancy_count = defaultdict(lambda: defaultdict(int))
        for i in range(0, len(marker_gene_list)):
            marker_gene_i = marker_gene_list[i]
            genes_in_genomes_i = gene_count_table[marker_gene_i]

            for j in range(i + 1, len(marker_genes)):
                marker_gene_j = marker_gene_list[j]
                genes_in_genomes_j = gene_count_table[marker_gene_j]

                for genome_id in genes_in_genomes_i:
                    if genome_id in genes_in_genomes_j:
                        if genes_in_genomes_i[genome_id].intersection(genes_in_genomes_j[genome_id]):
                            redundancy_count[marker_gene_i][marker_gene_j] += 1

        # Identify HMMs consistently hitting the same gene across genomes.
        #
        # Note that the following sets of redundant families is at least possible:
        #  X,Y
        #  X,Z
        #
        # It is unclear what to do in this case since, in general, we would expect
        # Y to also be redundant with Z. The following always resolves each redundant
        # pair in order giving preference to TIGRFAMs and HMMs with lower numbers. This
        # will NOT always result in the largest possible set of HMMs (i.e., Y, Z may
        # be removed when one could just remove X), but this seems fair since it is unclear
        # how to resolve such situations.
        hmms_to_remove = set()
        for marker_gene_i in redundancy_count:
            for marker_gene_j, count in redundancy_count[marker_gene_i].items():
                if count > redundancy:
                    if marker_gene_i in hmms_to_remove or marker_gene_j in hmms_to_remove:
                        # marker gene from this redundant pair is already marked for removal
                        continue

                    # preferentially discard PFAM models
                    if 'PF' in marker_gene_i and not 'PF' in marker_gene_j:
                        hmms_to_remove.add(marker_gene_i)
                        fout.write('%s\t%s\n' % (marker_gene_j, marker_gene_i))
                    elif not 'PF' in marker_gene_i and 'PF' in marker_gene_j:
                        hmms_to_remove.add(marker_gene_j)
                        fout.write('%s\t%s\n' % (marker_gene_i, marker_gene_j))
                    elif 'PF' in marker_gene_i and 'PF' in marker_gene_j:
                        # take Pfam model with lowest number as these tend
                        # to encode better known protein families
                        pfam_num_i = marker_gene_i.replace('PF', '')
                        pfam_num_i = int(pfam_num_i[0:pfam_num_i.find('.')])
                        pfam_num_j = marker_gene_j.replace('PF', '')
                        pfam_num_j = int(pfam_num_j[0:pfam_num_j.find('.')])

                        if pfam_num_i > pfam_num_j:
                            hmms_to_remove.add(marker_gene_i)
                            fout.write('%s\t%s\n' % (marker_gene_j, marker_gene_i))
                        else:
                            hmms_to_remove.add(marker_gene_j)
                            fout.write('%s\t%s\n' % (marker_gene_i, marker_gene_j))
                    else:
                        # take TIGRFAMs model with lowest number as these
                        # tend to be more universal and/or to encode better
                        # known protein families
                        tigr_num_i = int(marker_gene_i.replace('TIGR', ''))
                        tigr_num_j = int(marker_gene_j.replace('TIGR', ''))
                        if tigr_num_i > tigr_num_j:
                            hmms_to_remove.add(marker_gene_i)
                            fout.write('%s\t%s\n' % (marker_gene_j, marker_gene_i))
                        else:
                            hmms_to_remove.add(marker_gene_j)
                            fout.write('%s\t%s\n' % (marker_gene_i, marker_gene_j))

        fout.close()

        return hmms_to_remove

    def _fetch_marker_models(self, marker_genes, output_model_dir):
        """Save PFAM and TIGRFAM marker genes into individual model files.

        Parameters
        ----------
        marker_genes : iterable
            Marker genes to process.
        output_model_dir : str
            Directory to store HMM models.
        """

        marker_id_to_name = {}
        for line in open(self.pfam_model_file):
            if 'NAME' in line:
                name = line.split()[1].rstrip()
            elif 'ACC' in line:
                acc = line.split()[1].rstrip()
                marker_id_to_name[acc] = name

        for marker_id in marker_genes:
            if 'PF' in marker_id:
                os.system('hmmfetch ' + self.pfam_model_file + ' ' + marker_id_to_name[marker_id] + ' > ' + os.path.join(output_model_dir, marker_id + '.hmm'))
            else:
                model_file = os.path.join(self.tigrfams_model_dir, marker_id + '.HMM')
                os.system('hmmfetch ' + model_file + ' ' + marker_id + ' > ' + os.path.join(output_model_dir, marker_id + '.hmm'))

    def identify_marker_genes(self, ingroup_file,
                            ubiquity_threshold, single_copy_threshold, redundancy,
                            valid_marker_genes,
                            output_msa_dir, output_model_dir):
        """Identify ubiquitous, single-copy marker genes.

        Parameters
        ----------
        ingroup_file : str
            File specifying unique ids of ingroup genomes.
        ubiquity_threshold : float
            Threshold for defining ubiquity marker genes.
        single_copy_threshold : float
            Threshold for defining a single-copy marker gene.
        redundancy : float
            Threshold for declaring HMMs redundant.
        valid_marker_genes : iterable
            Restrict marker set to genes within this set.
        output_msa_dir : str
            Directory to store multiple sequence alignment of marker genes.
        output_model_dir : str
            Directory to store HMMs of marker genes.
        """

        # read directory for each genome
        genome_dirs = read_genome_dir_file(self.genome_dir_file)

        # read genomes within the ingroup
        ncbi_genome_ids, user_genome_ids = read_genome_id_file(ingroup_file)
        genome_ids = ncbi_genome_ids.union(user_genome_ids)
        self.logger.info('Ingroup genomes: %d' % len(genome_ids))
        self.logger.info('NCBI genomes: %d' % len(ncbi_genome_ids))
        self.logger.info('User genomes: %d' % len(user_genome_ids))

        # identify marker genes
        self.logger.info('Identifying marker genes.')
        gene_stats_file = os.path.join(output_model_dir, '..', 'gene_stats.all.tsv')
        gene_count_table = self._gene_count_table(genome_ids, genome_dirs)
        marker_gene_stats = self._marker_genes(genome_ids, gene_count_table, ubiquity_threshold, single_copy_threshold, gene_stats_file)

        # with open('tmp_marker_gene_list', 'wb') as f:
        #    pickle.dump(marker_gene_stats, f)

        # with open('tmp_marker_gene_list', 'rb') as f:
        #    marker_gene_stats = pickle.load(f)

        marker_genes = set(marker_gene_stats.keys())
        if valid_marker_genes:
            self.logger.info('Restricting %d identified markers to specified set of valid markers.' % len(marker_genes))
            marker_genes = marker_genes.intersection(valid_marker_genes)
        self.logger.info('Identified ubiquitous, single-copy marker genes: %d' % len(marker_genes))

        redundancy_out_file = os.path.join(output_model_dir, '..', 'redundant_markers.tsv')
        redundancy = redundancy * len(genome_ids)
        redundant_hmms = self._identify_redundant_hmms(marker_genes, gene_count_table, redundancy, redundancy_out_file)
        marker_genes = marker_genes - redundant_hmms
        self.logger.info('Marker genes identified as redundant between TIGRFAM and Pfam: %d' % len(redundant_hmms))
        self.logger.info('Remaining ubiquitous, single-copy marker genes: %d' % len(marker_genes))

        # get HMM for each marker gene
        self.logger.info('Fetching HMM for each marker genes.')
        self._fetch_marker_models(marker_genes, output_model_dir)

        # align gene sequences
        align_markers = AlignMarkers(self.cpus)
        align_markers.run(genome_ids, genome_dirs, marker_genes, False, output_msa_dir, output_model_dir)

        return len(genome_ids), len(ncbi_genome_ids), len(user_genome_ids), genome_ids, marker_gene_stats, marker_genes

    def infer_gene_trees(self, msa_dir, output_dir, extension):
        """Infer gene trees.

        Parameters
        ----------
        msa_dir : str
            Directory containing multiple sequence alignment of marker genes.
        output_dir : str
            Directory to store gene trees.
        extension : str
            Extension of multiple sequence alignment files.
        """

        files = os.listdir(msa_dir)
        msa_files = []
        for f in files:
            if f.endswith(extension):
                msa_file = os.path.join(msa_dir, f)
                msa_files.append(msa_file)

                fin = open(msa_file)
                data = fin.readlines()
                fin.close()

                fout = open(msa_file, 'w')
                for line in data:
                    if line[0] != '>':
                        # remove trailing star
                        if line[-1] == '*':
                            line = line[0:-1]
                    fout.write(line)
                fout.close()

        fasttree = FastTree(multithreaded=False)
        fasttree.parallel_run(msa_files, 'prot', 'wag', output_dir, self.cpus)

        # create gene tree without gene ids for visualization in ARB
        for msa_file in msa_files:
            tree_filename = ntpath.basename(msa_file)
            tree_prefix = tree_filename[0:tree_filename.find('.')]

            if tree_prefix.startswith('PF'):
                # patch up output file for Pfam trees
                old_tree_prefix = tree_prefix
                tree_prefix = '.'.join(tree_filename.split('.')[0:2])
                shutil.move(os.path.join(output_dir, old_tree_prefix + '.tree'),
                                os.path.join(output_dir, tree_prefix + '.tree'))

            gene_tree_file = os.path.join(output_dir, tree_prefix + '.tree')
            gene_tree = dendropy.Tree.get_from_path(gene_tree_file, schema='newick', rooting='force-unrooted', preserve_underscores=True)

            # rename nodes to contain only genome id
            for node in gene_tree.leaf_nodes():
                genome_id = node.taxon.label.split(DefaultValues.SEQ_CONCAT_CHAR)[0]
                node.taxon.label = genome_id

            output_tree_file = os.path.join(output_dir, tree_prefix + '.genome_ids.tree')
            gene_tree.write_to_path(output_tree_file, schema='newick', suppress_rooting=True, unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import logging

from biolib.common import make_sure_path_exists
from biolib.misc.time_keeper import TimeKeeper
from biolib.external.hmmer import HmmModelParser
from biolib.external.fasttree import FastTree

from genometreetk.common import (read_genome_id_file,
                                    read_genome_dir_file,
                                    read_marker_id_file,
                                    create_concatenated_alignment)
from genometreetk.markers.align_markers import AlignMarkers


class InferWorkflow(object):
    """Infer phylogenetic tree from concatenated marker genes."""

    def __init__(self, genome_dir_file, pfam_model_file, tigrfams_model_dir, cpus):
        """Initialization.

        Parameters
        ----------
        genome_dir_file : str
            File specifying directory for each genome.
        pfam_model_file : str
            File containing Pfam HMMs.
        tigrfams_model_dir : str
            Directory containing TIGRFAMs HMMs.
        cpus : int
            Number of cpus to use.
        """

        self.logger = logging.getLogger()

        self.genome_dir_file = genome_dir_file
        self.pfam_model_file = pfam_model_file
        self.tigrfams_model_dir = tigrfams_model_dir

        self.cpus = cpus

    def _fetch_marker_models(self, marker_genes, hmm_model_out, hmm_info_out, output_model_dir):
        """Save PFAM and TIGRFAM marker genes into a single HMM model file.

        Parameters
        ----------
        marker_genes : iterable
            Marker genes to fetch.
        hmm_model_out : str
            File to containing phylogenetically informative HMMs.
        hmm_info_out : str
            File to contain information about HMMs.
        output_model_dir : str
            Directory to write individual HMM model files.
        """

        marker_id_to_name = {}
        for line in open(self.pfam_model_file):
            if 'NAME' in line:
                name = line.split()[1].rstrip()
            elif 'ACC' in line:
                acc = line.split()[1].rstrip()
                marker_id_to_name[acc] = name

        fout_model = open(hmm_model_out, 'w')
        for marker_id in marker_genes:
            output_model_file = os.path.join(output_model_dir, marker_id + '.hmm')
            if 'PF' in marker_id:
                os.system('hmmfetch ' + self.pfam_model_file + ' ' + marker_id_to_name[marker_id] + ' > ' + output_model_file)
            else:
                input_model_file = os.path.join(self.tigrfams_model_dir, marker_id + '.HMM')
                os.system('hmmfetch ' + input_model_file + ' ' + marker_id + ' > ' + output_model_file)

            # write model to file
            for line in open(output_model_file):
                fout_model.write(line)

        fout_model.close()

        self.logger.info('    HMM models written to: ' + hmm_model_out)

        # read HMM model metadata
        hmm_model_parse = HmmModelParser(hmm_model_out)
        hmm_models = hmm_model_parse.models()

        fout_info = open(hmm_info_out, 'w')
        fout_info.write('Model Accession\tName\tDescription\tLength\n')
        for model in hmm_models.values():
            fout_info.write('%s\t%s\t%s\t%s\n' % (model.acc, model.name, model.desc, model.leng))
        fout_info.close()

        self.logger.info('    HMM information written to: ' + hmm_info_out)

    def run(self, genome_id_file,
                    marker_id_file,
                    model,
                    output_dir):
        """Identify phylogenetic tree.

        Parameters
        ----------
        genome_id_file : str
            File specifying unique ids of genomes to include in tree.
        marker_id_file : str
            File specifying unique ids of marker genes  to use for inference.
        model : str ['wag' or 'jtt']
            Model of evolution to use.
        output_dir : str
            Directory to store results.
        """

        time_keeper = TimeKeeper()

        output_alignment_dir = os.path.join(output_dir, 'alignments')
        make_sure_path_exists(output_alignment_dir)

        output_model_dir = os.path.join(output_dir, 'hmm_models')
        make_sure_path_exists(output_model_dir)

        # read directory for each genome
        genome_dirs = read_genome_dir_file(self.genome_dir_file)

        # read genomes within the ingroup
        ncbi_genome_ids, user_genome_ids = read_genome_id_file(genome_id_file)
        genome_ids = ncbi_genome_ids.union(user_genome_ids)
        self.logger.info('Inferring tree for %d genomes.' % len(genome_ids))
        self.logger.info('NCBI genomes: %d' % len(ncbi_genome_ids))
        self.logger.info('User genomes: %d' % len(user_genome_ids))

        # get marker genes
        self.logger.info('Reading marker genes.')
        marker_genes = read_marker_id_file(marker_id_file)
        self.logger.info('Read %d marker genes.' % len(marker_genes))

        # gather all single-copy HMMs into a single model file
        hmm_model_out = os.path.join(output_dir, 'phylo.hmm')
        hmm_info_out = os.path.join(output_dir, 'phylo.tsv')
        self.logger.info('Generating marker gene HMM model files.')
        self._fetch_marker_models(marker_genes, hmm_model_out, hmm_info_out, output_model_dir)

        # align gene sequences
        align_markers = AlignMarkers(self.cpus)
        align_markers.run(genome_ids, genome_dirs, marker_genes, True, output_alignment_dir, output_model_dir)

        # create concatenated alignment file
        self.logger.info('Concatenating alignments.')
        concatenated_alignment_file = os.path.join(output_dir, 'concatenated_alignment.faa')
        marker_file = os.path.join(output_dir, 'concatenated_markers.tsv')
        create_concatenated_alignment(genome_ids, marker_genes, output_alignment_dir, concatenated_alignment_file, marker_file)

        # create concatenated genome tree
        self.logger.info('Inferring concatenated genome tree.')
        concatenated_tree = os.path.join(output_dir, 'concatenated.tree')
        concatenated_tree_log = os.path.join(output_dir, 'concatenated.tree.log')
        log_file = os.path.join(output_dir, 'fasttree.log')
        fast_tree = FastTree(multithreaded=True)
        fast_tree.run(concatenated_alignment_file, 'prot', model, concatenated_tree, concatenated_tree_log, log_file)

        # generate summary report
        report_out = os.path.join(output_dir, 'infer_workflow.log')
        fout = open(report_out, 'w')
        fout.write('[infer]\n')
        fout.write('Genome Id file: %s\n' % genome_id_file)
        fout.write('Marker Id file: %s\n' % marker_id_file)
        fout.write('Model of evolution: %s\n' % model)
        fout.write(time_keeper.get_time_stamp())
        fout.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import csv
import sys
from collections import defaultdict, namedtuple

import biolib.seq_io as seq_io
from biolib.taxonomy import Taxonomy

from genometreetk.default_values import DefaultValues
from genometreetk.aai import aai_thresholds


# make sure large CSV files can be read
csv.field_size_limit(sys.maxsize)


def parse_genome_path(genome_path_file):
    """Determine path to genomic FASTA file for each genome."""
    # get path to genome FASTA files
        
    genome_files = {}
    for line in open(genome_path_file):
        gid, genome_path = line.strip().split('\t')
        accession = os.path.basename(os.path.normpath(genome_path))
        genome_files[gid] = os.path.join(genome_path, accession + '_genomic.fna')
        
    return genome_files
        

def binomial_species(taxonomy):
    """Get binomial, including Candidatus, species names in NCBI taxonomy."""
    
    binomial_names = defaultdict(set)
    for gid, taxa in taxonomy.items():
        species = taxa[6]
        if species == 's__':
            continue
            
        if any(c.isdigit() for c in species):
            continue
        
        is_candidatus = False
        if 'Candidatus ' in species:
            species = species.replace('Candidatus ', '')
            is_candidatus = True
            
        tokens = species[3:].split()
        if len(tokens) != 2:
            continue
            
        genus, specific = tokens

        if is_candidatus:
            species = species.replace('s__', 's__Candidatus ')
            
        if genus.istitle() and all(c.islower() for c in specific):
            binomial_names[species].add(gid)
        
    return binomial_names

    
def genome_species_assignments(ncbi_taxonomy):
    """Get species assignment for each genome."""
    
    ncbi_species = binomial_species(ncbi_taxonomy)
    gid_to_species = {}
    for sp in ncbi_species:
        for gid in ncbi_species[sp]:
            gid_to_species[gid] = sp
            
    return gid_to_species
        
    
def canonical_species_name(species_name):
    """Get canonical species name from GTDB or NCBI species name."""
    
    if species_name == 's__':
        return species_name
    
    species_name = species_name.replace('Candidatus ', '')
    prefix = species_name[0:3]
    full_name = species_name[3:]
    genus, species = full_name.split(' ')[0:2]
    
    underscore_pos = genus.rfind('_')
    if underscore_pos != -1:
        genus = genus[0:underscore_pos]
        
    underscore_pos = species.rfind('_')
    if underscore_pos != -1:
        species = species[0:underscore_pos]
        
    genus = genus.replace('[','').replace(']','')
    species = species.replace('[','').replace(']','')
        
    return prefix + genus + ' ' + species
    
    
def filter_genomes(metadata_file,
                    min_comp,
                    max_cont,
                    min_quality, 
                    max_contigs, 
                    min_N50, 
                    max_ambiguous, 
                    max_gap_length):
    """Indentify genomes passing filtering criteria."""
    
    genome_ids = set()

    csv_reader = csv.reader(open(metadata_file, 'rb'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            bHeader = False

            genome_index = row.index('accession')
            comp_index = row.index('checkm_completeness')
            cont_index = row.index('checkm_contamination')
            contig_count_index = row.index('contig_count')
            n50_scaffolds_index = row.index('n50_scaffolds')
            ambiguous_bases_index = row.index('ambiguous_bases')
            total_gap_length_index = row.index('total_gap_length')
            
        else:
            genome_id = row[genome_index]
            
            comp = float(row[comp_index])
            cont = float(row[cont_index])
            quality = comp - 5*cont
            
            contig_count = int(row[contig_count_index])
            n50_scaffolds = int(row[n50_scaffolds_index])
            ambiguous_bases = int(row[ambiguous_bases_index])
            total_gap_length = int(row[total_gap_length_index])
            
            if comp >= min_comp and cont <= max_cont and quality >= min_quality:
                if contig_count <= max_contigs and n50_scaffolds >= min_N50:
                    if ambiguous_bases <= max_ambiguous and total_gap_length <= max_gap_length:
                        genome_ids.add(genome_id)

    return genome_ids


def predict_bacteria(genome_id, bac_seqs, ar_seqs):
    """Check that domain assignment agrees with GTDB and NCBI taxonomy.
    
    Parameters
    ----------
    genome_id : str
        Genome of interest.
    bac_seqs : dict
        Bacterial sequences for all genomes.
    ar_seqs : dict
        Archaea sequences for all genomes.
    rep_is_bacteria : boolean
        Flag indicating if genome was classified as a Bacteria (True) or  Archaea (False).
        
    Returns
    -------
    boolean
        True if predicted domain is Bacteria, else False for Archaea.
    """
    rep_bac_seq = bac_seqs[genome_id]
    rep_ar_seq = ar_seqs[genome_id]
    
    per_bac_aa = float(len(rep_bac_seq) - rep_bac_seq.count('-')) / len(rep_bac_seq)
    per_ar_aa = float(len(rep_ar_seq) - rep_ar_seq.count('-')) / len(rep_ar_seq)
    
    is_bac = per_bac_aa >= per_ar_aa
    
    return is_bac, per_bac_aa, per_ar_aa
    
    
def reassign_representative(cur_representative_id,
                                cur_aai,
                                new_representative_id,
                                new_aai,
                                trusted_user_genomes):
    """Determines best representative genome.

    Genomes are preferentially assigned to representatives based on
    source repository (RefSeq => GenBank => Trusted User => User) and AAI.
    """
    
    source_order = {'R': 0,  # RefSeq
                    'G': 1,  # GenBank
                    'U': 2}  # User

    if not cur_representative_id:
        # no currently assigned representative
        return new_representative_id, new_aai

    cur_source = source_order[cur_representative_id[0]]
    new_source = source_order[new_representative_id[0]]
    if new_representative_id in trusted_user_genomes:
        new_source = source_order['G']
    
    if new_source < cur_source:
        # give preference to genome source
        return new_representative_id, new_aai
    elif new_source == cur_source:
        # for the same source, find the representative with the highest AAI
        if new_aai > cur_aai:
            return new_representative_id, new_aai

    return cur_representative_id, cur_aai

def assign_rep(rep_id, 
                    genome_id,
                    rep_is_bacteria, 
                    genome_is_bacteria,
                    bac_seqs, 
                    ar_seqs,
                    species,
                    gtdb_taxonomy,
                    genome_aa_count,
                    trusted_user_genomes,
                    aai_threshold,
                    min_matches,
                    assigned_representative, 
                    cur_aai):
    """Detering if genome should be assigned to a representative."""

    # do not cluster genomes from different predicted domains
    if rep_is_bacteria != genome_is_bacteria:
        return assigned_representative, cur_aai
        
    # do not cluster genomes from different named species
    rep_species = species.get(rep_id, None)
    genome_species = species.get(genome_id, None)
    if rep_species and genome_species and rep_species != genome_species:
        return assigned_representative, cur_aai

    # do not cluster NCBI or trusted User genomes with User representatives
    if rep_id.startswith('U_') and rep_id not in trusted_user_genomes: 
        if not genome_id.startswith('U__') or genome_id in trusted_user_genomes:
            return assigned_representative, cur_aai
            
    # do not cluster genomes from different named groups
    skip = False
    for rep_taxon, taxon in zip(gtdb_taxonomy[rep_id], gtdb_taxonomy[genome_id]):
        rep_taxon = rep_taxon[3:]
        taxon = taxon[3:]
        if rep_taxon and taxon and rep_taxon != taxon:
            skip = True
            break
    if skip:
        return assigned_representative, cur_aai
            
    if rep_is_bacteria:
        rep_seq = bac_seqs[rep_id]
        genome_seq = bac_seqs[genome_id]
    else:
        rep_seq = ar_seqs[rep_id]
        genome_seq = ar_seqs[genome_id]
        
    max_mismatches = (1.0 - cur_aai) * genome_aa_count
    aai = aai_thresholds(rep_seq, genome_seq, max_mismatches, min_matches)
    if aai > aai_threshold:  
        assigned_representative, cur_aai = reassign_representative(assigned_representative,
                                                                            cur_aai,
                                                                            rep_id,
                                                                            aai,
                                                                            trusted_user_genomes)
        
    return assigned_representative, cur_aai
    
        
def species_label(gtdb_taxonomy, ncbi_taxonomy, ncbi_organism_name):
    """Determine 'best' species label for each genome.

    Currently, this is just being set to the species label in the
    GTDB taxonomy. In theory, the NCBI taxonomy and organism name
    could also be consulted. However, since the GTDB taxonomy redefines
    some species this might be problematic so isn't currently being
    done.

    Parameters
    ----------
    gtdb_taxonomy : d[assembly_accession] -> [d__, ..., s__]
        GTDB taxonomy of each genome.
    ncbi_taxonomy : d[assembly_accession] -> [d__, ..., s__]
        NCBI taxonomy of each genome.
    ncbi_organism_name : d[assembly_accession] -> name
        NCBI organism name of each genome.

    Return
    ------
    dict : d[assembly_accession] -> species name
        Species name of each genome.
    """

    taxonomy = Taxonomy()

    species = {}
    species_index = Taxonomy.rank_index['s__']
    for genome_id, taxa in gtdb_taxonomy.items():
        sp = taxa[species_index]
        if sp != 's__':
            species[genome_id] = sp

    if False:   # do not consider NCBI information as
                # it may conflict with GTDB information
                # in unwanted ways
        
        for genome_id, taxa in ncbi_taxonomy.items():
            if genome_id in species:
                continue

            sp = taxa[species_index]
            sp = taxonomy.extract_valid_species_name(sp)
            if sp:
                species[genome_id] = sp

        for genome_id, sp in ncbi_organism_name.items():
            if genome_id in species:
                continue

            sp = taxonomy.extract_valid_species_name(sp)
            if sp:
                species[genome_id] = sp

    return species

    
def read_gtdb_metadata(metadata_file, fields):
    """Parse genome quality from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes in CSV file.
    fields : iterable
        Fields  to read.

    Return
    ------
    dict : d[genome_id] -> namedtuple
        Value for fields indicted by genome IDs.
    """

    gtdb_metadata = namedtuple('gtdb_metadata', ' '.join(fields))
    m = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')

        genome_index = headers.index('accession')

        indices = []
        for field in fields:
            indices.append(headers.index(field))

        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]

            values = []
            for i in indices:
                # save values as floats or strings
                v = line_split[i]
                try:
                    values.append(float(v))
                except ValueError:
                    if v is None or v == '' or v == 'none':
                        values.append(None)
                    elif v == 'f' or v.lower() == 'false':
                        values.append(False)
                    elif v == 't' or v.lower() == 'true':
                        values.append(True)
                    else:
                        values.append(v)
            m[genome_id] = gtdb_metadata._make(values)

    return m


def read_gtdb_phylum(metadata_file):
    """Parse GTDB phylum information from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> phyla
    """

    genome_phyla = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        phylum_index = headers.index('gtdb_phylum')

        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            genome_phyla[genome_id] = line_split[phylum_index]

    return genome_phyla


def read_gtdb_taxonomy(metadata_file):
    """Parse GTDB taxonomy from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> taxonomy list
    """

    taxonomy = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        taxonomy_index = headers.index('gtdb_taxonomy')
        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            taxa_str = line_split[taxonomy_index].strip()

            if taxa_str and taxa_str != 'none':
                taxonomy[genome_id] = map(str.strip, taxa_str.split(';'))
            else:
                taxonomy[genome_id] = list(Taxonomy.rank_prefixes)

    return taxonomy
    
def read_gtdb_representative(metadata_file):
    """Parse GTDB representative from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> True or False
    """

    gtdb_reps = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        gtdb_representative_index = headers.index('gtdb_representative')

        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            is_rep = (line_split[gtdb_representative_index] == 't')
            gtdb_reps[genome_id] = is_rep

    return gtdb_reps


def read_gtdb_ncbi_taxonomy(metadata_file, species_exception_file):
    """Parse NCBI taxonomy from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> taxonomy list
    """

    taxonomy = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        taxonomy_index = headers.index('ncbi_taxonomy')
        
        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            taxa_str = line_split[taxonomy_index].strip()
            taxa_str = taxa_str.replace('Candidatus ', '')

            if taxa_str and taxa_str != 'none':
                taxonomy[genome_id] = map(str.strip, taxa_str.split(';'))
            else:
                taxonomy[genome_id] = list(Taxonomy.rank_prefixes)
    
    ncbi_update_count = 0
    if species_exception_file:
        for line in open(species_exception_file):
            line_split = line.strip().split('\t')
            gid = line_split[0]
            if gid == 'Genome ID':
                continue
                
            sp = line_split[1].replace('Candidatus ', '')
            if gid not in taxonomy:
                print('Genome in species exception list not defined at NCBI: %s' % gid)
                sys.exit(-1)
                
            if not sp.startswith('s__'):
                sp = 's__' + sp
                
            taxonomy[gid][6] = sp
            ncbi_update_count += 1

    return taxonomy, ncbi_update_count


def read_gtdb_ncbi_organism_name(metadata_file):
    """Parse NCBI organism name from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    dict : d[genome_id] -> organism name
        Organism name of each genome.
    """

    d = {}

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        organism_name_index = headers.index('ncbi_organism_name')

        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            organism_name = line_split[organism_name_index].strip()

            if organism_name:
                d[genome_id] = organism_name

    return d


def read_gtdb_ncbi_type_strain(metadata_file):
    """Parse NCBI type strain from GTDB metadata.

    Parameters
    ----------
    metadata_file : str
        Metadata for all genomes.

    Return
    ------
    set
        Set of genomes marked as type strains by NCBI.
    """

    type_strains = set()

    with open(metadata_file) as f:
        headers = f.readline().strip().split('\t')
        genome_index = headers.index('accession')
        type_strain_index = headers.index('ncbi_type_strain')

        for line in f:
            line_split = line.strip().split('\t')
            genome_id = line_split[genome_index]
            if bool(line_split[type_strain_index]):
                type_strains.add(genome_id)

    return type_strains


def read_tree_model(reportFile):
    for line in open(reportFile):
        if 'Model of evolution:' in line:
            modelStr = line[line.find(':') + 1:].strip()

    return modelStr


def read_genome_dir_file(genome_dir_file):
    """Read genome directories from file.

    Parameters
    ----------
    genome_dir_file : str
        File to read.

    Returns
    -------
    dict : d[genome_id] -> directory
        Directory for each genome.
    """

    # read directory for each genome
    genome_dirs = {}
    for line in open(genome_dir_file):
        line_split = line.split('\t')
        genome_dirs[line_split[0]] = line_split[1].strip()

    return genome_dirs


def read_marker_id_file(marker_id_file):
    """Read marker ids from file.

    Parameters
    ----------
    marker_id_file : str
        File to read.

    Returns
    -------
    set
        Marker ids.
    """

    marker_genes = set()
    for line in open(marker_id_file):
        if line[0] == '#':
            continue

        marker_genes.add(line.split('\t')[0].strip())

    return marker_genes


def read_genome_id_file(genome_id_file):
    """Read genome ids from file.

    Parameters
    ----------
    genome_file : str
        File to read.

    Returns
    -------
    set
        NCBI genome ids.
    set
        User genome ids.
    """

    ncbi_genome_ids = set()
    user_genome_ids = set()
    for line in open(genome_id_file):
        if line[0] == '#':
            continue

        if '\t' in line:
            genome_id = line.split('\t')[0].strip()
        else:
            genome_id = line.split()[0].strip()

        if genome_id.startswith('U_'):
            user_genome_ids.add(genome_id)
        else:
            ncbi_genome_ids.add(genome_id)

    return ncbi_genome_ids, user_genome_ids


def create_concatenated_alignment(genome_ids,
                                   marker_genes,
                                   alignment_dir,
                                   concatenated_alignment_file,
                                   marker_file):
    """Create concatenated multiple sequence alignment for all genomes.

    Parameters
    ----------
    genome_ids : iterable
        Genomes of interest.
    marker_genes : iterable
        Unique ids of marker genes.
    alignment_dir : str
        Directory containing multiple sequence alignments.
    concatenated_alignment_file : str
        File to containing concatenated alignment.
    marker_file : str
        File indicating length of each marker in the alignment.
    """

    # Read alignment files. Some genomes may have multiple
    # copies of a marker gene in which case the last one
    # is arbitrarily taken. This is acceptable as all genes
    # are already screen to be conspecific.
    alignments = defaultdict(dict)
    marker_length = {}
    for mg in marker_genes:
        f = mg + '.aln.masked.faa'
        seqs = seq_io.read_fasta(os.path.join(alignment_dir, f))

        for seq_id, seq in seqs.items():
            genome_id = seq_id[0:seq_id.find(DefaultValues.SEQ_CONCAT_CHAR)]

            alignments[mg][genome_id] = seq

            marker_length[mg] = len(seq)

    # create marker file
    fout = open(marker_file, 'w')
    for mg in marker_genes:
        fout.write('%s\t%s\t%s\t%d\n' % (mg, mg, mg, marker_length[mg]))
    fout.close()

    # create concatenated alignment
    concatenated_seqs = {}
    for mg in marker_genes:
        seqs = alignments[mg]

        for genome_id in genome_ids:
            if genome_id in seqs:
                # append alignment
                concatenated_seqs[genome_id] = concatenated_seqs.get(genome_id, '') + seqs[genome_id]
            else:
                # missing gene
                concatenated_seqs[genome_id] = concatenated_seqs.get(genome_id, '') + '-' * marker_length[mg]

    # save concatenated alignment
    seq_io.write_fasta(concatenated_seqs, concatenated_alignment_file)




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import itertools

def aai_thresholds(seq1, seq2, max_mismatches, min_matches):
    """Calculate AAI between sequences.

    Mismatches are only calculate across
    positions where both sequences have
    an amino acid. The max_mismatches
    threshold is used to quickly stop
    comparisons between divergent
    sequences.

    Parameters
    ----------
    seq1 : str
        First sequence.
    seq2 : float
        Second sequence.
    max_mismatches : int
        Maximum allowed mismatches between sequences.
    min_matches : int
        Minimum required matches between sequences.

    Returns
    -------
    float
        AAI between sequences, or 0 if criteria is not meet.
    """

    mismatches = 0
    matches = 0
    for c1, c2 in itertools.izip(seq1, seq2):
        if c1 == '-' or c2 == '-':
            continue
        elif c1 != c2:
            mismatches += 1
            if mismatches > max_mismatches:
                return 0
        else:
            matches += 1
            
    if matches < min_matches:
        return 0

    aai = float(matches) / (matches + mismatches)
    return aai


def aai(seq1, seq2, threshold):
    """Determine AAI between sequences.

    The identify is only calculate across
    positions where both sequences have
    an amino acid. A threshold value is
    used to allow quick identification
    of highly similar sequences.

    Parameters
    ----------
    seq1 : str
        First sequence.
    seq2 : float
        Second sequence.
    threshold : float
        Minimum AAI required for reporting.

    Returns
    -------
    float
        AAI between sequences if it is greater than the
        specified threshold, else None.
    """

    assert len(seq1) == len(seq2)

    max_mismatches = (1.0 - threshold) * len(seq1)

    mismatches = 0
    matches = 0
    for c1, c2 in itertools.izip(seq1, seq2):
        if c1 == '-' or c2 == '-':
            continue
        elif c1 != c2:
            mismatches += 1
            if mismatches >= max_mismatches:
                return None
        else:
            matches += 1

    aai = float(matches) / max(1, (matches + mismatches))
    if aai < threshold:
        return None

    return aai


def aai_test(seq1, seq2, threshold):
    """Test AAI between sequences.

    The identify is only calculate across
    positions where both sequences have
    an amino acid. A threshold value is
    used to allow quick identification
    of highly similar sequences.

    Parameters
    ----------
    seq1 : str
        First sequence.
    seq2 : float
        Second sequence.
    threshold : float
        Minimum AAI required for reporting.

    Returns
    -------
    bool
        True if AAI >= threshold, else False.
    """

    if aai(seq1, seq2, threshold):
        return True

    return False



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import logging

import biolib.seq_io as seq_io
from biolib.common import remove_extension
from biolib.external.fasttree import FastTree
from biolib.external.blast import Blast
from biolib.taxonomy import Taxonomy

import genometreetk.ncbi as ncbi
from genometreetk.common import (read_gtdb_metadata,
                                    read_genome_dir_file,
                                    read_gtdb_taxonomy)
                                    

class RNA_Workflow(object):
    """Infer RNA gene trees."""

    def __init__(self, cpus):
        """Initialization.

        Parameters
        ----------
        cpus : int
            Maximum number of CPUs to use.
        """

        self.logger = logging.getLogger()
        self.cpus = cpus
        
    def _dump_seqs(self, 
                    genomic_file, 
                    gtdb_taxonomy,
                    genomes_of_interest,
                    prefix,
                    min_ar_gene_len,
                    min_bac_gene_len,
                    min_contig_len,
                    output_prefix,
                    output_dir):
            
        fout_ar_summary = open(os.path.join(output_dir, output_prefix + '_ar.tsv'), 'w')
        fout_ar_fna = open(os.path.join(output_dir, output_prefix + '_ar.fna'), 'w')
        fout_ar_taxonmy = open(os.path.join(output_dir, output_prefix + '_ar_taxonomy.tsv'), 'w')
        
        fout_bac_summary = open(os.path.join(output_dir, output_prefix + '_bac.tsv'), 'w')
        fout_bac_fna = open(os.path.join(output_dir, output_prefix + '_bac.fna'), 'w')
        fout_bac_taxonmy = open(os.path.join(output_dir, output_prefix + '_bac_taxonomy.tsv'), 'w')
        
        write_header = True
        total_seq = 0
        for line in open(genomic_file):
            gid, genome_path = [t.strip() for t in line.split()]
            if gid.startswith('GCA_'):
                gid = 'GB_' + gid
            elif gid.startswith('GCF_'):
                gid = 'RS_' + gid
            
            if genomes_of_interest and gid not in genomes_of_interest:
                continue

            if 'd__Archaea' in gtdb_taxonomy[gid]:
                fout_summary = fout_ar_summary
                fout_fna = fout_ar_fna
                fout_taxonomy = fout_ar_taxonmy
                min_gene_len = min_ar_gene_len
            else:
                fout_summary = fout_bac_summary
                fout_fna = fout_bac_fna
                fout_taxonomy = fout_bac_taxonmy
                min_gene_len = min_bac_gene_len

            # extract sequences
            hmm_summary = os.path.join(genome_path, prefix + '.hmm_summary.tsv')
            if not os.path.exists(hmm_summary):
                continue
                
            seqs = seq_io.read(os.path.join(genome_path, prefix + '.fna'))
            gene_count = 0
            with open(hmm_summary) as f:
                header = f.readline()
                if write_header:
                    write_header = False
                    fout_summary.write('%s\t%s' % ('Gene ID', header))
                
                for line in f:
                    line_split = line.strip().split('\t')
                    gene_id = line_split[0]
                    gene_len = int(line_split[5])
                    contig_len = int(line_split[-1])
                    
                    if gene_len >= min_gene_len and contig_len >= min_contig_len:
                        unique_gene_id = '%s~gene_%s' % (gid, gene_count)
                        fout_summary.write('%s\t%s' % (unique_gene_id, line))
                        fout_fna.write('>%s [%s]\n' % (unique_gene_id, gene_id))
                        fout_fna.write(seqs[gene_id] + '\n')
                        fout_taxonomy.write('%s\t%s\n' % (unique_gene_id, '; '.join(gtdb_taxonomy[gid])))
                        
                        gene_count += 1
                        total_seq += 1
                
        fout_ar_summary.close()
        fout_ar_fna.close()
        fout_ar_taxonmy.close()
        
        fout_bac_summary.close()
        fout_bac_fna.close()
        fout_bac_taxonmy.close()

        self.logger.info('Wrote %d sequences.' % total_seq)
        
    def dump(self, 
                genomic_file,
                gtdb_taxonomy,
                min_5S_len,
                min_16S_ar_len,
                min_16S_bac_len,
                min_23S_len,
                min_contig_len,
                include_user,
                genome_list,
                output_dir):
        """Dump 5S, 16S, and 23S sequences to files."""
        
        if include_user:
            self.logger.warning('User genomes not currently supported.')
            sys.exit(-1)
            
        gtdb_taxonomy = Taxonomy().read(gtdb_taxonomy)
            
        genomes_of_interest = set()
        if genome_list:
            for line in open(genome_list):
                line_split = line.strip().split('\t')
                gid = line_split[0]
                if gid.startswith('GCA_'):
                    gid = 'GB_' + gid
                elif gid.startswith('GCF_'):
                    gid = 'RS_' + gid
                genomes_of_interest.add(gid)

            self.logger.info('Restricting gene dump to %d genomes.' % len(genomes_of_interest))
            
        self.logger.info('Dumping 5S sequences.')
        self._dump_seqs(genomic_file,
                        gtdb_taxonomy,
                        genomes_of_interest,
                        'lsu_5S/lsu_5S', 
                        min_5S_len, 
                        min_5S_len,
                        min_contig_len,
                        'lsu_5s',
                        output_dir)
        
        self.logger.info('Dumping 16S sequences.')
        self._dump_seqs(genomic_file,
                        gtdb_taxonomy,
                        genomes_of_interest,
                        'rna_silva/ssu', 
                        min_16S_ar_len, 
                        min_16S_bac_len,
                        min_contig_len,
                        'ssu',
                        output_dir)
        
        self.logger.info('Dumping 23S sequences.')
        self._dump_seqs(genomic_file,
                        gtdb_taxonomy,
                        genomes_of_interest,
                        'rna_silva/lsu_23S', 
                        min_23S_len, 
                        min_23S_len,
                        min_contig_len,
                        'lsu_23s',
                        output_dir)

    def _get_rna_seqs(self,
                        rna_name,
                        rna_file,
                        min_rna_length,
                        min_scaffold_length,
                        gtdb_taxonomy,
                        genomes_to_consider,
                        output_dir):
        """Get 16S sequences from genomes.
        
        Assumes the FASTA header has the format:
            <genome_id>~<seq_id> <gtdb_taxonomy> <rna_gene_length> <contig_length>

        Parameters
        ----------
        rna_name : str
            Name of rRNA gene.
        rna_file : str
            File with rRNA gene sequences in FASTA format.
        min_rna_length : int
            Minimum required length of rRNA gene sequences.
        min_scaffold_length : int
            Minimum required length of scaffold containing rRNE gene sequence.
        gtdb_taxonomy : list
            GTDB taxonomy for each genome assembly.
        genomes_to_consider : iterable
            IDs of genomes to obtain 16S sequences.
        output_dir : str
            Desired output directory.

        Returns
        -------
        str
            Name of file with 16S sequences in FASTA format.
        """

        rna_output_file = os.path.join(output_dir, '%s.fna' % rna_name)
        fout = open(rna_output_file, 'w')
        filtered_rna_len = 0
        filtered_scaffold_len = 0
        genomes_with_seq = set()
        for seq_id, seq, annotation in seq_io.read_seq(rna_file, keep_annotation=True):
            genome_id, contig_id = seq_id.split('~', 1)
            if genome_id not in genomes_to_consider:
                continue
                
            if len(seq) < min_rna_length:
                filtered_rna_len += 1
                continue

            scaffold_len = int(annotation.split(' ')[-1].strip())
            if scaffold_len < min_scaffold_length:
                filtered_scaffold_len += 1
                continue
                
            genomes_with_seq.add(genome_id)
            
            fout.write('>%s %s\n' % (genome_id, annotation))
            fout.write(seq + '\n')
            
        fout.close()

        self.logger.info('Filtered %d rRNA gene sequence based on rRNA gene length <%d.' % (filtered_rna_len, min_rna_length))
        self.logger.info('Filtered %d rRNA gene sequence based on scaffold length <%d.' % (filtered_scaffold_len, min_scaffold_length))
        self.logger.info('Identified %d genomes with a valid rRNA gene sequence.' % len(genomes_with_seq))

        return rna_output_file

    def _trim_seqs(self, input_msa, output_msa, remove_identical=False, min_per_taxa=0.5, min_bp=1000):
        """Trim ends of sequences.

        input_msa : str
            File with MSA to trim.
        output_msa : str
            New file with trimmed MSA.
        remove_identical : boolean
            Flag indicating if identical sequence should be removed.
        min_per_taxa : float [0, 1.0]
            Minimum required taxa to retain leading and trailing columns.
        min_bp : int
            Minimum required length to retain sequence.
        """

        # read seqs
        seqs = seq_io.read_fasta(input_msa)

        # filter identical seqs
        identical_seqs = set()
        if remove_identical:
            self.logger.info('Filtering identical sequences.')

            seq_ids = seqs.keys()
            for i in range(0, len(seq_ids)):
                seq_id_I = seq_ids[i]

                if seq_id_I in identical_seqs:
                    continue

                for j in range(i + 1, len(seqIds)):
                    seq_id_J = seq_ids[j]
                    if seqs[seq_id_I] == seqs[seq_id_J]:
                      self.logger.info('Seq %s and %s are identical.' % (seq_id_I, seq_id_J))
                      identical_seqs.add(seq_id_J)

            self.logger.info('Identified %d of %d sequences as identical.' % (len(identical_seqs), len(seqs)))

        # trim start and end columns to consensus alignment
        first_char = []
        last_char = []
        for seq_id, seq in seqs.items():
            if seq_id in identical_seqs:
                continue

            for i, ch in enumerate(seq):
                if ch != '.' and ch != '-':
                    first_char.append(i)
                    break

            for i in range(len(seq) - 1, -1, -1):
                if seq[i] != '.' and seq[i] != '-':
                    last_char.append(i)
                    break

        first_char.sort()
        last_char.sort(reverse=True)

        trim_index = int((len(seqs) * min_per_taxa))

        start = first_char[trim_index]
        end = last_char[trim_index]

        self.logger.info('Trimming seqs from %d to %d leaving a %dbp length alignment.' % (start, end, end - start + 1))

        short_seq_file = output_msa + '.short'
        fout = open(output_msa, 'w')
        fout_short = open(short_seq_file, 'w')
        num_filtered_seq = 0
        for seq_id, seq in seqs.items():
            if seq_id in identical_seqs:
                continue

            valid_bp = 0
            for i in range(start, min(len(seq), end + 1)):
                ch = seq[i]
                if ch != '.' and ch != '-':
                    valid_bp += 1

            if valid_bp >= min_bp:
                fout.write('>' + seq_id + '\n')
                fout.write(seq[start:end + 1] + '\n')
            else:
                self.logger.info('Filtering seq %s with %d of %d (%.1f%%) aligned bases.' % (seq_id, valid_bp, (end - start + 1), valid_bp * 100.0 / (end - start + 1)))
                num_filtered_seq += 1
                fout_short.write('>' + seq_id + '\n')
                fout_short.write(seq[start:end + 1] + '\n')

        fout.close()
        fout_short.close()

        self.logger.info('Filtered %d of %d sequences due to length.' % (num_filtered_seq, len(seqs) - len(identical_seqs)))
        self.logger.info('Short sequence written to: %s' % short_seq_file)

    def _tax_filter(self, ssu_output_file, taxonomy, output_dir):
        """Identify sequence to filter based on taxonomy of best BLAST hit.

        """
        
        extant_taxa = Taxonomy().extant_taxa(taxonomy)

        tax_filter_dir = os.path.join(output_dir, 'tax_filter')
        if not os.path.exists(tax_filter_dir):
            os.makedirs(tax_filter_dir)
        
        blast = Blast(self.cpus)
        self.logger.info('Creating BLASTN database.')
        blast.create_blastn_db(ssu_output_file)

        self.logger.info('Performing reciprocal BLAST to identify sequences with incongruent taxonomies.')
        
        blast_table = os.path.join(tax_filter_dir, 'blastn.tsv')
        blast.blastn(ssu_output_file,
                     ssu_output_file,
                     blast_table,
                     evalue=1e-10,
                     max_matches=2,
                     output_fmt='custom',
                     task='blastn')

        filter = set()
        fout = open(os.path.join(tax_filter_dir, 'filtered_seqs.tsv'), 'w')
        fout.write('Query ID\tQuery Taxonomy\tSubject ID\tSubject Taxonomy\tPerc. Identity\tAlign. Length\tMismatch Rank\tNo. Query Genomes\tNo. Subject Genomes\n')
        for hit in blast.read_hit(blast_table, table_fmt='custom'):
            if hit.query_id == hit.subject_id:
                # ignore self hits
                continue

            if hit.alignment_len > 800:
                query_genome_id = hit.query_id.split('~', 1)[0]
                subject_genome_id = hit.subject_id.split('~', 1)[0]
                    
                # require a (very lenient) percent identity of threshold from Yarza et al., 2014
                if hit.query_id not in filter and hit.perc_identity >= 82: # order
                    rank_index = Taxonomy.rank_labels.index('order')
                    query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                    subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                    if query_taxa and subject_taxa and query_taxa != subject_taxa:
                        filter.add(hit.query_id)
                        fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                ';'.join(taxonomy[query_genome_id]),
                                                                hit.subject_id,
                                                                ';'.join(taxonomy[subject_genome_id]),
                                                                hit.perc_identity,
                                                                hit.alignment_len,
                                                                'Order',
                                                                len(extant_taxa['o__' + query_taxa]),
                                                                len(extant_taxa['o__' + subject_taxa])))
                
                if False:
                    if hit.perc_identity >= 75: # phylum
                        rank_index = Taxonomy.rank_labels.index('phylum')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Phylum',
                                                                    len(extant_taxa['p__' + query_taxa]),
                                                                    len(extant_taxa['p__' + subject_taxa])))
                                                                    
                    if hit.perc_identity >= 78.5: # class
                        rank_index = Taxonomy.rank_labels.index('class')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Class',
                                                                    len(extant_taxa['c__' + query_taxa]),
                                                                    len(extant_taxa['c__' + subject_taxa])))
                    
                    if hit.query_id not in filter and hit.perc_identity >= 82: # order
                        rank_index = Taxonomy.rank_labels.index('order')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Order',
                                                                    len(extant_taxa['o__' + query_taxa]),
                                                                    len(extant_taxa['o__' + subject_taxa])))
                                                                    
                    if hit.query_id not in filter and hit.perc_identity >= 86.5: # family
                        rank_index = Taxonomy.rank_labels.index('family')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Family',
                                                                    len(extant_taxa['f__' + query_taxa]),
                                                                    len(extant_taxa['f__' + subject_taxa])))
                                                                    
                    if hit.query_id not in filter and hit.perc_identity >= 94.5: # genus
                        rank_index = Taxonomy.rank_labels.index('genus')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Genus',
                                                                    len(extant_taxa['g__' + query_taxa]),
                                                                    len(extant_taxa['g__' + subject_taxa])))

                    if hit.query_id not in filter and hit.perc_identity >= 99: # species
                        rank_index = Taxonomy.rank_labels.index('species')
                        query_taxa = taxonomy[query_genome_id][rank_index][3:].strip()
                        subject_taxa = taxonomy[subject_genome_id][rank_index][3:].strip()
                        if query_taxa and subject_taxa and query_taxa != subject_taxa:
                            filter.add(hit.query_id)
                            fout.write('%s\t%s\t%s\t%s\t%.2f\t%d\t%s\t%d\t%d\n' % (hit.query_id, 
                                                                    ';'.join(taxonomy[query_genome_id]),
                                                                    hit.subject_id,
                                                                    ';'.join(taxonomy[subject_genome_id]),
                                                                    hit.perc_identity,
                                                                    hit.alignment_len,
                                                                    'Species',
                                                                    len(extant_taxa['s__' + query_taxa]),
                                                                    len(extant_taxa['s__' + subject_taxa])))

        fout.close()
        
        return filter

    def run(self, rna_name,
                    gtdb_metadata_file, 
                    rna_file, 
                    min_rna_length,
                    min_scaffold_length,
                    min_quality,
                    max_contigs,
                    min_N50,
                    tax_filter,
                    genome_list,
                    output_dir,
                    align_method='ssu_align'):
        """Infer rRNA gene tree spanning select GTDB genomes.

        Parameters
        ----------
        rna_name : str
            Name of rRNA gene.
        gtdb_metadata_file : str
            File specifying GTDB metadata for each genome.
        rna_file : str
            File with rRNA gene sequences in FASTA format.
        min_rna_length : int
            Minimum required length of rRNA gene sequences.
        min_scaffold_length : int
            Minimum required length of scaffold containing rRNA gene sequence.
        min_quality : float [0, 100]
            Minimum genome quality for a genome to be include in tree.
        max_contigs : int
            Maximum number of contigs to include genome.
        min_N50 : int
            Minimum N50 to include genome.
        tax_filter : boolean
            Filter sequences based on incongruent taxonomy classification.
        genome_list : str
            Explicit list of genomes to use (ignores --ncbi_rep_only and --user_genomes).
        output_dir : str
            Directory to store results
        """
                
        if rna_name not in ['ssu', 'lsu']:
            self.logger.error('Unrecognized rRNA gene type: %s' % rna_name)
            sys.exit(-1)

        genome_metadata = read_gtdb_metadata(gtdb_metadata_file, ['checkm_completeness',
                                                                      'checkm_contamination',
                                                                      'scaffold_count',
                                                                      'n50_scaffolds',
                                                                      'organism_name',
                                                                      'gtdb_representative'])

        gtdb_taxonomy = read_gtdb_taxonomy(gtdb_metadata_file)

        user_genomes = set()
        uba_genomes = set()
        ncbi_genomes = set()
        rep_genomes = set()
        for genome_id in genome_metadata:
            org_name = str(genome_metadata[genome_id][4])
            if genome_id.startswith('U_'):
                if '(UBA' in org_name:
                    uba_genomes.add(genome_id)
                else:
                    user_genomes.add(genome_id)
            elif genome_id.startswith('RS_') or genome_id.startswith('GB_'):
                ncbi_genomes.add(genome_id)
            else:
                self.logger.warning('Unrecognized genome prefix: %s' % genome_id)
                
            rep = genome_metadata[genome_id][5] == 't'
            if rep:
                rep_genomes.add(genome_id)
                
        self.logger.info('Initially considering %d genomes (%d NCBI, %d UBA, %d User).' % (len(genome_metadata), 
                                                                                            len(ncbi_genomes), 
                                                                                            len(uba_genomes), 
                                                                                            len(user_genomes)))
        self.logger.info('Identified %d representative genomes.' % len(rep_genomes))
        
        # get genomes specified in genome list by user
        genomes_to_consider = set()
        if genome_list:
            for line in open(genome_list):
                gid = line.rstrip().split('\t')[0]
                if gid.startswith('RS_') or gid.startswith('GB_') or gid.startswith('U_'):
                    genomes_to_consider.add(gid)
            self.logger.info('Restricting genomes to the %d in the genome list.' % len(genomes_to_consider))
        else:
            # filter genomes based on quality and database source
            self.logger.info('Filtering genomes based on specified critieria.')
            self.logger.info('Filtering on minimum quality <%d.' % min_quality)
            self.logger.info('Filtering on number of contigs >%d.' % max_contigs)
            self.logger.info('Filtering on scaffold N50 <%d.' % min_N50)
            
            new_genomes_to_consider = []
            filtered_genomes = 0
            gt = 0
            gq = 0
            sc = 0
            n50 = 0
            for genome_id in genome_metadata:                
                if genome_id not in rep_genomes:
                    gt += 1
                    filtered_genomes += 1
                    continue
                    
                if genome_id not in ncbi_genomes and genome_id not in uba_genomes:
                    gt += 1
                    filtered_genomes += 1
                    continue
                    
                comp, cont, scaffold_count, n50_contigs, _org_name, _rep = genome_metadata[genome_id]
                q = float(comp) - 5*float(cont)
                if q < min_quality or int(scaffold_count) > max_contigs or int(n50_contigs) < min_N50:
                    if q < min_quality:
                        gq += 1
                        
                    if int(scaffold_count) > max_contigs:
                        sc += 1
                        
                    if int(n50_contigs) < min_N50:
                        n50 += 1
                
                    filtered_genomes += 1
                    continue
                    
                new_genomes_to_consider.append(genome_id)

            genomes_to_consider = new_genomes_to_consider
            self.logger.info('Filtered %d genomes (%d on genome type, %d on genome quality, %d on number of contigs, %d on N50).' % (filtered_genomes, gt, gq, sc, n50))
            self.logger.info('Considering %d genomes after filtering.' % len(genomes_to_consider))
            
        # limit taxonomy to genomes being considered
        cur_gtdb_taxonomy = {}
        for gid in genomes_to_consider:
            cur_gtdb_taxonomy[gid] = gtdb_taxonomy[gid]

        # get rRNA gene sequences for each genome
        rna_output_file = self._get_rna_seqs(rna_name,
                                                rna_file,
                                                min_rna_length,
                                                min_scaffold_length,
                                                cur_gtdb_taxonomy,
                                                genomes_to_consider,
                                                output_dir)

        # identify erroneous rRNA gene sequences
        if tax_filter:
            self.logger.info('Filtering sequences with incongruent taxonomy strings.')
            filter = self._tax_filter(rna_output_file, cur_gtdb_taxonomy, output_dir)

            self.logger.info('Filtered %d sequences.' % len(filter))
            if len(filter) > 0:
                rna_filtered_output = os.path.join(output_dir, 'gtdb_%s.tax_filter.fna' % rna_name)
                fout = open(rna_filtered_output, 'w')
                for seq_id, seq, annotation in seq_io.read_seq(rna_output_file, keep_annotation=True):
                    if seq_id not in filter:
                        fout.write('>' + seq_id + ' ' + annotation + '\n')
                        fout.write(seq + '\n')
                fout.close()

                rna_output_file = rna_filtered_output

        # align sequences with ssu-align or mothur
        if rna_name == 'ssu':
            if align_method == 'ssu_align':
                self.logger.info('Aligning sequences with ssu-align.')
                align_dir = os.path.join(output_dir, '%s_align' % rna_name)
                os.system('ssu-align --dna %s %s' % (rna_output_file, align_dir))
                os.system('ssu-mask --afa %s' % align_dir)
            elif align_method == 'mothur':
                self.logger.info('Aligning sequences with mothur.')
                align_dir = os.path.join(output_dir, 'mothur')
                if not os.path.exists(align_dir):
                    os.makedirs(align_dir)

                mothur_cmd = 'mothur "#set.dir(output=%s, blastdir=/srv/sw/Mothur/1.39.5)' % align_dir
                mothur_cmd += '; align.seqs(candidate=%s, template=/srv/db/mothur/silva_128/silva.seed_v128.align, search=blast, flip=t, processors=%d)' % (rna_output_file, self.cpus)
                input_prefix = remove_extension(rna_output_file)
                align_file = os.path.join(align_dir, input_prefix + '.align')
                mothur_cmd += '; filter.seqs(fasta=%s, hard=/srv/db/mothur/silva_128/Lane1349.silva.filter, processors=%d);"' % (align_file, self.cpus)
                os.system(mothur_cmd)
                input_msa = os.path.join(align_dir, input_prefix + '.filter.fasta')
        elif rna_name == 'lsu':
            self.logger.info('Aligning sequences with ssu-align.')
            align_dir = os.path.join(output_dir, '%s_align' % rna_name)
            if not os.path.exists(align_dir):
                os.makedirs(align_dir)
                
            os.system('esl-sfetch --index %s' % rna_output_file)
                  
            # search fo sequences using domain-specific LSU HMMs
            for domain in ['archaea', 'bacteria', 'eukaryote']:
                self.logger.info('Matching LSU rRNA genes to %s-specific HMM.' % domain)
                table_out = os.path.join(align_dir, 'cmsearch.%s.%s.tblout' % (rna_name, domain))
                cm_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cm_files')            
                cm_file = os.path.join(cm_dir, 'lsu_%s.cm' % domain)
                log_file = os.path.join(align_dir, 'cmsearch.%s.%s.out' % (rna_name, domain))
                os.system('cmsearch --hmmonly --cpu %d --noali --tblout %s %s %s > %s' % (self.cpus, table_out, cm_file, rna_output_file, log_file))
                
            # identify top hits for each domain
            self.logger.info('Identifying best domain-specific HMM for each LSU rRNA gene.')
            top_hits = {}
            for domain in ['archaea', 'bacteria', 'eukaryote']:
                table_out = os.path.join(align_dir, 'cmsearch.%s.%s.tblout' % (rna_name, domain))
                for line in open(table_out):
                    if line[0] == '#':
                        continue
                        
                    line_split = line.split()
                    seq_id = line_split[0]
                    start_seq = int(line_split[7])
                    end_seq = int(line_split[8])
                    bitscore = float(line_split[14])

                    prev_bitscore = top_hits.get(seq_id, [None, 0, 0, 0, 0])[4]
                    if bitscore > prev_bitscore:
                        top_hits[seq_id] = [domain, seq_id, start_seq, end_seq, bitscore]
              
            # create MSA for each bacteria and archaea
            for domain in ['archaea', 'bacteria']:
                # creat file of top hits
                top_hits_out = os.path.join(align_dir, 'top_hits.%s.%s.tsv' % (rna_name, domain))
                fout = open(top_hits_out, 'w')
                num_hits = 0
                for top_domain, seq_id, start_seq, end_seq, bitscore in top_hits.values():
                    if top_domain == domain:
                        fout.write('%s\t%d\t%d\%f\n' % (seq_id, start_seq, end_seq, bitscore))
                        num_hits += 1
                fout.close()
                
                # align top hits
                self.logger.info('Creating MSA for %s LSU rRNA genes (%d sequences).' % (domain, num_hits))
                
                if num_hits > 0:
                    seq_file = os.path.join(align_dir, 'cmsearch.%s.%s.fna' % (rna_name, domain))
                    os.system("grep -v '^#' %s | awk '{print $1, $2, $3, $1}' | esl-sfetch -Cf %s - > %s" % (top_hits_out, rna_output_file, seq_file))
                    
                    align_file = os.path.join(align_dir, 'cmalign.%s.%s.stk' % (rna_name, domain))
                    os.system('cmalign --dnaout --outformat Pfam %s %s > %s' % (cm_file, seq_file, align_file))
                    
                    masked_file = os.path.join(align_dir, 'cmalign.%s.%s.mask.afa' % (rna_name, domain))
                    os.system('esl-alimask -p --outformat AFA %s > %s' % (align_file, masked_file))
            
        # trim sequences and infer tree
        if align_method == 'ssu_align':
            for domain in ['archaea', 'bacteria']:
                if rna_name == 'ssu':
                    input_msa = os.path.join(align_dir, 'ssu_align.' + domain + '.mask.afa')
                elif rna_name == 'lsu':
                    input_msa = os.path.join(align_dir, 'cmalign.%s.%s.mask.afa' % (rna_name, domain))
                    
                if not os.path.exists(input_msa):
                    continue
                    
                trimmed_msa = os.path.join(output_dir, domain + '.trimmed.fna')
                self._trim_seqs(input_msa, trimmed_msa)

                # infer tree
                self.logger.info('Inferring tree for %s genes.' % domain)
                output_tree = os.path.join(output_dir, domain + '.tree')
                os.system('FastTreeMP -nosupport -nt -gtr -gamma %s > %s' % (trimmed_msa, output_tree))
        elif align_method == 'mothur':
            trimmed_msa = os.path.join(output_dir, input_prefix + '.trimmed.fna')
            self._trim_seqs(input_msa, trimmed_msa)

            # infer tree
            self.logger.info('Inferring tree for %s genes.')
            output_tree = os.path.join(output_dir, input_prefix + '.tree')
            os.system('FastTreeMP -nosupport -nt -gtr -gamma %s > %s' % (trimmed_msa, output_tree))
            
    def combine(self, ssu_msa, ssu_tree, lsu_msa, lsu_tree, output_dir):
        """Infer 16S + 23S tree spanning GTDB genomes."""
        
        # identify common 16S and 23S sequences
        ssu_seqs = {}
        for seq_id, seq, annotation in seq_io.read_seq(ssu_msa, keep_annotation=True):
            genome_id = seq_id.split('~')[0]
            ssu_seqs[genome_id] = [seq, annotation]
        self.logger.info('Read %d SSU rRNA sequences.' % len(ssu_seqs))
            
        lsu_seqs = {}
        for seq_id, seq, annotation in seq_io.read_seq(lsu_msa, keep_annotation=True):
            genome_id = seq_id.split('~')[0]
            lsu_seqs[genome_id] = [seq, annotation]
        self.logger.info('Read %d LSU rRNA sequences.' % len(lsu_seqs))
              
        common_seqs = set(ssu_seqs.keys()).intersection(lsu_seqs.keys())
        self.logger.info('Identified %d sequences in common.' % len(common_seqs))
        
        # identify incongruent taxonomic order classifcations between trees
        self.logger.info('Identifying incongruent order-level taxonomic classifications between trees.')
        ssu_taxonomy = Taxonomy().read_from_tree(ssu_tree)
        lsu_taxonomy = Taxonomy().read_from_tree(lsu_tree)

        order_index = Taxonomy.rank_labels.index('order')

        seqs_to_filter = set()
        for seq_id in common_seqs:
            ssu_order = ssu_taxonomy.get(seq_id)[order_index][3:]
            lsu_order = lsu_taxonomy.get(seq_id)[order_index][3:]
            
            # remove designator of paraphyletic orders
            # (since in the concatenated tree this may be resolved)
            ssu_order = ssu_order.split('_')[0]
            lsu_order = lsu_order.split('_')[0]
            
            if ssu_order != lsu_order:
                seqs_to_filter.add(seq_id)
        
        self.logger.info('Identified %d sequences with incongruent classifcations.' % len(seqs_to_filter))
        common_seqs.difference_update(seqs_to_filter)
        
        # write out MSA
        concatenated_msa = os.path.join(output_dir, 'ssu_lsu_concatenated.fna')
        fout = open(concatenated_msa, 'w')
        for seq_id in common_seqs:
            fout.write('>%s %s %s\n' % (seq_id, ssu_seqs[seq_id][1], lsu_seqs[seq_id][1]))
            fout.write('%s%s\n' % (ssu_seqs[seq_id][0], lsu_seqs[seq_id][0]))
        fout.close()
        
        # infer tree
        output_tree = os.path.join(output_dir, 'ssu_lsu_concatenated.tree')
        os.system('FastTreeMP -nosupport -nt -gtr -gamma %s > %s' % (concatenated_msa, output_tree))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging
from collections import defaultdict

from math import floor

import dendropy

from biolib.common import is_float
from biolib.taxonomy import Taxonomy


class PhylogeneticDiversity():
    """Calculate phylogenetic diversity."""

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger()
        
    def _read_taxa_list(self, taxa_list):
        """Read taxa from file."""
        
        taxa = set()
        for line in open(taxa_list):
            taxa.add(line.strip().split('\t')[0])
            
        return taxa
        
    def _total_pd(self, tree):
        """Calculate PD over entire tree."""
        
        total_pd = 0
        total_taxa = 0
        for node in tree.preorder_node_iter():
            if node.parent_node is not None:
                total_pd += node.edge.length
                if node.is_leaf():
                    total_taxa += 1
                    
        return total_pd, total_taxa

    def _taxon_pd(self, tree, ingroup, outgroup):
        """Calculate phylogenetic gain of each ingroup taxon relative to outgroup."""

        pg_taxon = {}
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label in ingroup:
                # find first internal node containing an outgroup taxon
                pg = 0
                parent = leaf
                outgroup_taxon = 'None'
                while parent:
                    # check for outgroup taxon
                    stop = False
                    for tip in parent.leaf_iter():
                        if tip.taxon.label in outgroup:
                            outgroup_taxon = tip.taxon.label
                            if outgroup_taxon in ingroup:
                                outgroup_taxon += ' (one or more outgroup taxa are assigned to this ingroup taxon)'
                            stop = True
                            
                    if stop:
                        break
                    
                    pg += parent.edge.length
                    parent = parent.parent_node
                    
                pg_taxon[leaf.taxon.label] = [pg, outgroup_taxon]

        return pg_taxon
        
    def _taxa_pd(self, tree, taxa):
        """Calculate phylogenetic diversity of a set of taxa."""
        
        pd = 0
        visited_edges = set()
        for cur_node in taxa:
            while cur_node.parent_node is not None:
                if cur_node.edge in visited_edges:
                    break

                pd += cur_node.edge.length
                visited_edges.add(cur_node.edge)
                cur_node = cur_node.parent_node

        return pd

    def pd(self, tree, taxa_list, per_taxa_pg_file):
        """Calculate phylogenetic diversity of extant taxa."""

        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)
                                                                            
        # get total branch length of tree
        self.logger.info('Calculating total PD.')
        total_pd, total_taxa = self._total_pd(tree)
        self.logger.info('Total PD for %d taxa = %.2f' % (total_taxa, total_pd))

        # get PD of ingroup taxa
        self.logger.info('Calculating total PD for specified taxa.')
        ingroup = self._read_taxa_list(taxa_list)
        
        in_taxa = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label in ingroup:
                in_taxa.add(leaf)
        
        self.logger.info('Specified ingroup taxa: %d' % len(ingroup))
        self.logger.info('Ingroup taxa found in tree: %d' % len(in_taxa))

        # calculate PD for ingroup
        in_pd = self._taxa_pd(tree, in_taxa)

        # get PD of outgroup taxa
        outgroup = set()
        out_taxa = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label not in ingroup:
                out_taxa.add(leaf)
                outgroup.add(leaf.taxon.label)
                
        self.logger.info('Outgroup taxa found in tree: %d' % len(out_taxa))
        out_pd = self._taxa_pd(tree, out_taxa)

        # calculate PG of each ingroup taxon relative to outgroup in requested
        if per_taxa_pg_file:
            self.logger.info('Calculating PG of each ingroup taxon relative to outgroup.')
            pg_taxon = self._taxon_pd(tree, ingroup, outgroup)
            
            fout = open(per_taxa_pg_file, 'w')
            fout.write('Taxon\tPG\tPercent PG\tFirst outgroup taxon\n')
            for taxon, pg_stats in pg_taxon.items():
                pg, outgroup_taxon = pg_stats
                fout.write('%s\t%f\t%f\t%s\n' % (taxon, pg, pg * 100.0 / total_pd, outgroup_taxon))
            fout.close()
                   
        return total_pd, len(in_taxa), in_pd, len(out_taxa), out_pd
        
    def _clade_pd(self, tree, ingroup, outgroup):
        """Calculate PD for named clades."""
        
        pd = {}
        for node in tree.preorder_node_iter():
            if not node.label:
                continue

            taxon = None
            if ':' in node.label:
                _support, taxon = node.label.split(':')
            else:
                if not is_float(node.label):
                    taxon = node.label
                    
            if taxon:
                taxon_pd = 0
                taxon_count = 0
                in_taxon_pd = 0
                in_taxon_count = 0
                out_taxon_pd = 0
                out_taxon_count = 0
                for nn in node.postorder_iter():
                    if nn == node:
                        continue
                        
                    # check if group contains taxa from
                    # the ingroup and/or outgroup
                    ingroup_leaves = False
                    outgroup_leaves = False
                    for leaf in nn.leaf_iter():
                        genome_id = leaf.taxon.label
                        if genome_id in ingroup:
                            ingroup_leaves = True
                        
                        if genome_id in outgroup:
                            outgroup_leaves = True
                            
                    if ingroup_leaves:
                        in_taxon_pd += nn.edge.length

                    if outgroup_leaves:
                        out_taxon_pd += nn.edge.length
                        
                    if nn.is_leaf():
                        genome_id = nn.taxon.label
                        
                        if genome_id in ingroup:
                            in_taxon_count += 1
                        else:
                            out_taxon_count += 1
                            
                    taxon_pd += nn.edge.length

                pd[taxon] = [taxon_pd, in_taxon_pd, in_taxon_count, out_taxon_pd, out_taxon_count]

        return pd
        
    def pd_clade(self, decorated_tree, taxa_list, output_file):
        """Calculate phylogenetic diversity of named groups."""
        
        # calculate PD for entire tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(decorated_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)
 
        self.logger.info('Calculating total PD.')
        total_pd, total_taxa = self._total_pd(tree)
        
        # get ingroup and outgroup taxa
        ingroup = self._read_taxa_list(taxa_list)
        outgroup = set()
        in_tree = set()
        for leaf in tree.leaf_node_iter():
            if leaf.taxon.label not in ingroup:
                outgroup.add(leaf.taxon.label)
            if leaf.taxon.label in ingroup:
                in_tree.add(leaf.taxon.label)
                
        self.logger.info('Specified %d ingroup taxa.' % len(ingroup))
        self.logger.info('Identified %d ingroup taxa in tree.' % len(in_tree))
        self.logger.info('Identified %d outgroup taxa in tree.' % len(outgroup))

        # PD for named groups
        self.logger.info('Calculating PD for named clades.')
        pd_clade = self._clade_pd(tree, ingroup, outgroup)
        self.logger.info(' ... identified %d named clades.' % len(pd_clade))

        # report results
        fout = open(output_file, 'w')
        fout.write('Clade')
        fout.write('\tTaxa\tPD\tPercent PD')
        fout.write('\tOut Taxa\tOut PD\tOut Percent PD')
        fout.write('\tIn Taxa\tIn PD\tIn Percent PD')
        fout.write('\tIn PG\tIn Percent PG\n')
        
        ordered_taxa = Taxonomy().sort_taxa(pd_clade.keys())
        for taxon in ordered_taxa:
            taxon_pd, in_taxon_pd, in_taxon_count, out_taxon_pd, out_taxon_count = pd_clade[taxon]
            taxon_count = in_taxon_count + out_taxon_count
            in_taxon_pg = taxon_pd - out_taxon_pd
            
            taxon_pd = max(taxon_pd, 1e-9) # make sure PD is never exactly zero to avoid division errors
            
            row = taxon
            row += '\t%d\t%.2f\t%.2f' % (taxon_count, taxon_pd, taxon_pd * 100 / total_pd)
            row += '\t%d\t%.2f\t%.2f' % (out_taxon_count, out_taxon_pd, out_taxon_pd * 100 / taxon_pd)
            row += '\t%d\t%.2f\t%.2f' % (in_taxon_count, in_taxon_pd, in_taxon_pd * 100 / taxon_pd)
            row += '\t%.2f\t%.2f' % (in_taxon_pg, in_taxon_pg * 100 / taxon_pd)
            fout.write(row + '\n')
        fout.close()



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import csv
import logging
import random
from collections import defaultdict

import dendropy

from biolib.common import check_file_exists, make_sure_path_exists, is_float
from biolib.external.execute import check_dependencies
from biolib.taxonomy import Taxonomy
from biolib.newick import parse_label

from genometreetk.exceptions import GenomeTreeTkError
from genometreetk.rna_workflow import RNA_Workflow
from genometreetk.bootstrap import Bootstrap
from genometreetk.jackknife_markers import JackknifeMarkers
from genometreetk.jackknife_taxa import JackknifeTaxa
from genometreetk.combine_support import CombineSupport
from genometreetk.reroot_tree import RerootTree
from genometreetk.common import parse_genome_path, read_gtdb_metadata, read_gtdb_taxonomy
from genometreetk.phylogenetic_diversity import PhylogeneticDiversity
from genometreetk.arb import Arb
from genometreetk.derep_tree import DereplicateTree


csv.field_size_limit(sys.maxsize)

class OptionsParser():
    def __init__(self):
        """Initialization"""
        self.logger = logging.getLogger()

    def ssu_tree(self, options):
        """Infer 16S tree spanning GTDB genomes."""

        check_dependencies(['mothur', 'ssu-align', 'ssu-mask', 'FastTreeMP', 'blastn'])

        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.gtdb_ssu_file)
        make_sure_path_exists(options.output_dir)

        rna_workflow = RNA_Workflow(options.cpus)
        rna_workflow.run('ssu',
                            options.gtdb_metadata_file,
                            options.gtdb_ssu_file,
                            options.min_ssu_length,
                            options.min_scaffold_length,
                            options.min_quality,
                            options.max_contigs,
                            options.min_N50,
                            not options.disable_tax_filter,
                            options.genome_list,
                            options.output_dir,
                            options.align_method)

        self.logger.info('Results written to: %s' % options.output_dir)
        
    def lsu_tree(self, options):
        """Infer 23S tree spanning GTDB genomes."""

        check_dependencies(['esl-sfetch', 'cmsearch', 'cmalign', 'esl-alimask', 'FastTreeMP', 'blastn'])

        check_file_exists(options.gtdb_metadata_file)
        check_file_exists(options.gtdb_lsu_file)
        make_sure_path_exists(options.output_dir)

        rna_workflow = RNA_Workflow(options.cpus)
        rna_workflow.run('lsu',
                            options.gtdb_metadata_file,
                            options.gtdb_lsu_file,
                            options.min_lsu_length,
                            options.min_scaffold_length,
                            options.min_quality,
                            options.max_contigs,
                            options.min_N50,
                            not options.disable_tax_filter,
                            #options.reps_only,
                            #options.user_genomes,
                            options.genome_list,
                            options.output_dir)

        self.logger.info('Results written to: %s' % options.output_dir)
        
    def rna_tree(self, options):
        """Infer 16S + 23S tree spanning GTDB genomes."""

        check_dependencies(['FastTreeMP'])

        check_file_exists(options.ssu_msa)
        check_file_exists(options.ssu_tree)
        check_file_exists(options.lsu_msa)
        check_file_exists(options.lsu_tree)
        make_sure_path_exists(options.output_dir)

        rna_workflow = RNA_Workflow(options.cpus)
        rna_workflow.combine(options.ssu_msa,
                                options.ssu_tree,
                                options.lsu_msa,
                                options.lsu_tree,
                                options.output_dir)

        self.logger.info('Results written to: %s' % options.output_dir)
        
    def rna_dump(self, options):
        """Dump all 5S, 16S, and 23S sequences to files."""
        
        check_file_exists(options.genomic_file)
        make_sure_path_exists(options.output_dir)
        
        rna_workflow = RNA_Workflow(1)
        rna_workflow.dump(options.genomic_file,
                            options.gtdb_taxonomy,
                            options.min_5S_len,
                            options.min_16S_ar_len,
                            options.min_16S_bac_len,
                            options.min_23S_len,
                            options.min_contig_len,
                            options.include_user,
                            options.genome_list,
                            options.output_dir)
                            
        self.logger.info('Results written to: %s' % options.output_dir)
        
    def derep_tree(self, options):
        """Dereplicate tree."""
        
        check_file_exists(options.input_tree)
        check_file_exists(options.gtdb_metadata)
        check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)
        
        derep_tree = DereplicateTree()
        derep_tree.run(options.input_tree,
                        options.lineage_of_interest,
                        options.outgroup,
                        options.gtdb_metadata,
                        options.taxa_to_retain,
                        options.msa_file,
                        options.keep_unclassified,
                        options.output_dir)

    def bootstrap(self, options):
        """Bootstrap multiple sequence alignment."""

        check_file_exists(options.input_tree)
        if options.msa_file != 'NONE':
            check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        bootstrap = Bootstrap(options.cpus)
        output_tree = bootstrap.run(options.input_tree,
                                    options.msa_file,
                                    options.num_replicates,
                                    options.model,
                                    options.gamma,
                                    options.base_type,
                                    options.fraction,
                                    options.boot_dir,
                                    options.output_dir)

        self.logger.info('Bootstrapped tree written to: %s' % output_tree)

    def jk_markers(self, options):
        """Jackknife marker genes."""

        check_file_exists(options.input_tree)
        if options.msa_file != 'NONE':
            check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        jackknife_markers = JackknifeMarkers(options.cpus)
        output_tree = jackknife_markers.run(options.input_tree,
                                                options.msa_file,
                                                options.marker_info_file,
                                                options.mask_file,
                                                options.perc_markers,
                                                options.num_replicates,
                                                options.model,
                                                options.jk_dir,
                                                options.output_dir)

        self.logger.info('Jackknifed marker tree written to: %s' % output_tree)

    def jk_taxa(self, options):
        """Jackknife taxa."""

        check_file_exists(options.input_tree)
        check_file_exists(options.msa_file)
        make_sure_path_exists(options.output_dir)

        jackknife_taxa = JackknifeTaxa(options.cpus)
        output_tree = jackknife_taxa.run(options.input_tree,
                                            options.msa_file,
                                            options.outgroup_ids,
                                            options.perc_taxa,
                                            options.num_replicates,
                                            options.model,
                                            options.output_dir)

        self.logger.info('Jackknifed taxa tree written to: %s' % output_tree)

    def combine(self, options):
        """Combine support values into a single tree."""

        combineSupport = CombineSupport()
        combineSupport.run(options.support_type,
                            options.bootstrap_tree,
                            options.jk_marker_tree,
                            options.jk_taxa_tree,
                            options.output_tree)

    def support_wf(self, options):
        """"Perform entire tree support workflow."""

        self.bootstrap(options)
        self.jk_markers(options)
        self.jk_taxa(options)
        self.combine(options)

    def midpoint(self, options):
        """"Midpoint root tree."""

        reroot = RerootTree()
        reroot.midpoint(options.input_tree, options.output_tree)

    def outgroup(self, options):
        """Reroot tree with outgroup."""

        check_file_exists(options.taxonomy_file)

        self.logger.info('Identifying genomes from the specified outgroup.')
        outgroup = set()
        for genome_id, taxa in Taxonomy().read(options.taxonomy_file).items():
            if options.outgroup_taxon in taxa:
                outgroup.add(genome_id)
        self.logger.info('Identifying %d genomes in the outgroup.' % len(outgroup))

        reroot = RerootTree()
        reroot.root_with_outgroup(options.input_tree, options.output_tree, outgroup)

    def fill_ranks(self, options):
        """Ensure taxonomy strings contain all 7 canonical ranks."""

        check_file_exists(options.input_taxonomy)

        fout = open(options.output_taxonomy, 'w')
        taxonomy = Taxonomy()
        t = taxonomy.read(options.input_taxonomy)

        for genome_id, taxon_list in t.items():
            full_taxon_list = taxonomy.fill_missing_ranks(taxon_list)

            taxonomy_str = ';'.join(full_taxon_list)
            if not taxonomy.check_full(taxonomy_str):
                sys.exit(-1)

            fout.write('%s\t%s\n' % (genome_id, taxonomy_str))

        fout.close()

        self.logger.info('Revised taxonomy written to: %s' % options.output_taxonomy)

    def propagate(self, options):
        """Propagate labels to all genomes in a cluster."""

        check_file_exists(options.input_taxonomy)
        check_file_exists(options.metadata_file)

        # get representative genome information
        rep_metadata = read_gtdb_metadata(options.metadata_file, ['gtdb_representative',
                                                                  'gtdb_clustered_genomes'])
                                                                  
        taxonomy = Taxonomy()
        explict_tax = taxonomy.read(options.input_taxonomy)
        expanded_taxonomy = {}
        incongruent_count = 0
        for genome_id, taxon_list in explict_tax.items():
            taxonomy_str = ';'.join(taxon_list)

            # Propagate taxonomy strings if genome is a representatives. Also, determine
            # if genomes clustered together have compatible taxonomies. Note that a genome
            # may not have metadata as it is possible a User has removed a genome that is
            # in the provided taxonomy file.
            _rep_genome, clustered_genomes = rep_metadata.get(genome_id, (None, None))
            if clustered_genomes:  # genome is a representative
                clustered_genome_ids = clustered_genomes.split(';')

                # get taxonomy of all genomes in cluster with a specified taxonomy
                clustered_genome_tax = {}
                for cluster_genome_id in clustered_genome_ids:
                    if cluster_genome_id == genome_id:
                        continue

                    if cluster_genome_id not in rep_metadata:
                        continue  # genome is no longer in the GTDB so ignore it

                    if cluster_genome_id in explict_tax:
                        clustered_genome_tax[cluster_genome_id] = explict_tax[cluster_genome_id]

                # determine if representative and clustered genome taxonomy strings are congruent
                working_cluster_taxonomy = list(taxon_list)
                incongruent_with_rep = False
                for cluster_genome_id, cluster_tax in clustered_genome_tax.items():
                    if incongruent_with_rep:
                        working_cluster_taxonomy = list(taxon_list)  # default to rep taxonomy
                        break

                    for r in range(0, len(Taxonomy.rank_prefixes)):
                        if cluster_tax[r] == Taxonomy.rank_prefixes[r]:
                            break  # no more taxonomy information to consider

                        if cluster_tax[r] != taxon_list[r]:
                            if taxon_list[r] == Taxonomy.rank_prefixes[r]:
                                # clustered genome has a more specific taxonomy string which
                                # should be propagate to the representative if all clustered
                                # genomes are in agreement
                                if working_cluster_taxonomy[r] == Taxonomy.rank_prefixes[r]:
                                    # make taxonomy more specific based on genomes in cluster
                                    working_cluster_taxonomy[r] = cluster_tax[r]
                                elif working_cluster_taxonomy[r] != cluster_tax[r]:
                                    # not all genomes agree on the assignment of this rank so leave it unspecified
                                    working_cluster_taxonomy[r] = Taxonomy.rank_prefixes[r]
                                    break
                            else:
                                # genomes in cluster have incongruent taxonomies so defer to representative
                                self.logger.warning("Genomes in cluster have incongruent taxonomies.")
                                self.logger.warning("Representative %s: %s" % (genome_id, taxonomy_str))
                                self.logger.warning("Clustered genome %s: %s" % (cluster_genome_id, ';'.join(cluster_tax)))
                                self.logger.warning("Deferring to taxonomy specified for representative.")

                                incongruent_count += 1
                                incongruent_with_rep = True
                                break

                cluster_taxonomy_str = ';'.join(working_cluster_taxonomy)

                # assign taxonomy to representative and all genomes in the cluster
                expanded_taxonomy[genome_id] = cluster_taxonomy_str
                for cluster_genome_id in clustered_genome_ids:
                    expanded_taxonomy[cluster_genome_id] = cluster_taxonomy_str
            else:
                if genome_id in expanded_taxonomy:
                    # genome has already been assigned a taxonomy based on its representative
                    pass
                else:
                    # genome is a singleton
                    expanded_taxonomy[genome_id] = taxonomy_str


        self.logger.info('Identified %d clusters with incongruent taxonomies.' % incongruent_count)

        fout = open(options.output_taxonomy, 'w')
        for genome_id, taxonomy_str in expanded_taxonomy.items():
            fout.write('%s\t%s\n' % (genome_id, taxonomy_str))
        fout.close()

        self.logger.info('Taxonomy written to: %s' % options.output_taxonomy)

    def strip(self, options):
        """Remove taxonomic labels from tree."""

        check_file_exists(options.input_tree)

        outgroup_in_tree = set()
        tree = dendropy.Tree.get_from_path(options.input_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        for node in tree.internal_nodes():
            if node.label:
                if ':' in node.label:
                    support, _taxa = node.label.split(':')
                    node.label = support
                else:
                    try:
                        # if number if a float (or int) treat
                        # it as a support value
                        f = float(node.label)
                    except ValueError:
                        node.label = None

        tree.write_to_path(options.output_tree,
                            schema='newick',
                            suppress_rooting=True,
                            unquoted_underscores=True)

        self.logger.info('Stripped tree written to: %s' % options.output_tree)
        
    def rm_support(self, options):
        """Remove support values from tree."""

        check_file_exists(options.input_tree)

        outgroup_in_tree = set()
        tree = dendropy.Tree.get_from_path(options.input_tree,
                                            schema='newick',
                                            rooting='force-rooted',
                                            preserve_underscores=True)

        for node in tree.internal_nodes():
            if node.label:
                if ':' in node.label:
                    support, taxa = node.label.split(':')
                    node.label = taxa
                else:
                    try:
                        # if number if a float (or int) treat
                        # it as a support value
                        f = float(node.label)
                        node.label = None
                    except ValueError:
                        pass # keep other labels

        tree.write_to_path(options.output_tree,
                            schema='newick',
                            suppress_rooting=True,
                            unquoted_underscores=True)

        self.logger.info('Stripped tree written to: %s' % options.output_tree)
        
    def pull(self, options):
        """Create taxonomy file from a decorated tree."""

        check_file_exists(options.input_tree)

        if options.no_validation:
            tree = dendropy.Tree.get_from_path(options.input_tree, 
                                                schema='newick', 
                                                rooting="force-rooted", 
                                                preserve_underscores=True)

            taxonomy = {}
            for leaf in tree.leaf_node_iter():
                taxon_id = leaf.taxon.label
                
                node = leaf.parent_node
                taxa = []
                while node:
                    support, taxon, aux_info = parse_label(node.label)
                    if taxon:
                        for t in map(str.strip, taxon.split(';'))[::-1]:
                            taxa.append(t)
                    node = node.parent_node
                    
                taxonomy[taxon_id] = taxa[::-1]
        else:
            taxonomy = Taxonomy().read_from_tree(options.input_tree)
                                                
        Taxonomy().write(taxonomy, options.output_taxonomy)
            
        self.logger.info('Stripped tree written to: %s' % options.output_taxonomy)
        
    def append(self, options):
        """Append command"""
        
        check_file_exists(options.input_tree)
        check_file_exists(options.input_taxonomy)

        taxonomy = Taxonomy().read(options.input_taxonomy)

        tree = dendropy.Tree.get_from_path(options.input_tree, 
                                            schema='newick', 
                                            rooting='force-rooted', 
                                            preserve_underscores=True)
        
        for n in tree.leaf_node_iter():
            taxa_str = taxonomy.get(n.taxon.label, None)
            if taxa_str == None:
                self.logger.error('Taxonomy file does not contain an entry for %s.' % n.label)
                sys.exit(-1)
            n.taxon.label = n.taxon.label + '|' + '; '.join(taxonomy[n.taxon.label])

        tree.write_to_path(options.output_tree, 
                            schema='newick', 
                            suppress_rooting=True, 
                            unquoted_underscores=True)

        self.logger.info('Decorated tree written to: %s' % options.output_tree)
        
    def phylogenetic_diversity(self, options):
        """Calculate phylogenetic diversity of extant taxa."""
        
        check_file_exists(options.tree)
        check_file_exists(options.taxa_list)
        
        pd = PhylogeneticDiversity()
        rtn = pd.pd(options.tree, options.taxa_list, options.per_taxa_pg_file)
        total_pd, num_in_taxa, in_pd, num_out_taxa, out_pd = rtn
        total_taxa = num_in_taxa + num_out_taxa
        in_pg = total_pd - out_pd
                                            
        # report phylogenetic diversity (PD) and gain (PG)
        print ''
        print '\tNo. Taxa\tPD\tPercent PD'
        
        print '%s\t%d\t%.2f\t%.2f%%' % ('Full tree', total_taxa, total_pd, 100)
        
        print '%s\t%d\t%.2f\t%.3f%%' % ('Outgroup taxa (PD)',
                                            num_out_taxa,
                                            out_pd, 
                                            out_pd * 100 / total_pd)

        print '%s\t%d\t%.2f\t%.3f%%' % ('Ingroup taxa (PD)',
                                            num_in_taxa,
                                            in_pd, 
                                            (in_pd) * 100 / total_pd)   
                                        
        print '%s\t%d\t%.2f\t%.3f%%' % ('Ingroup taxa (PG)',
                                            num_in_taxa,
                                            in_pg, 
                                            in_pg * 100 / total_pd)
                  
    def phylogenetic_diversity_clade(self, options):
        """Calculate phylogenetic diversity of named groups."""

        check_file_exists(options.decorated_tree)
        
        pd = PhylogeneticDiversity()
        pd.pd_clade(options.decorated_tree, 
                    options.taxa_list, 
                    options.output_file)
        
    def arb_records(self, options):
        """Create an ARB records file from GTDB metadata."""

        check_file_exists(options.metadata_file)
        
        arb = Arb()
        arb.create_records(options.metadata_file, 
                            options.msa_file, 
                            options.taxonomy_file, 
                            options.genome_list, 
                            options.output_file)

    def parse_options(self, options):
        """Parse user options and call the correct pipeline(s)"""

        logging.basicConfig(format='', level=logging.INFO)

        check_dependencies(('FastTree', 'hmmsearch'))

        if options.subparser_name == 'ssu_tree':
            self.ssu_tree(options)
        elif options.subparser_name == 'lsu_tree':
            self.lsu_tree(options)
        elif options.subparser_name == 'rna_tree':
            self.rna_tree(options)
        elif options.subparser_name == 'rna_dump':
            self.rna_dump(options)
        elif options.subparser_name == 'derep_tree':
            self.derep_tree(options)
        elif options.subparser_name == 'bootstrap':
            self.bootstrap(options)
        elif options.subparser_name == 'jk_markers':
            self.jk_markers(options)
        elif options.subparser_name == 'jk_taxa':
            self.jk_taxa(options)
        elif options.subparser_name == 'combine':
            self.combine(options)
        elif options.subparser_name == 'midpoint':
            self.midpoint(options)
        elif options.subparser_name == 'outgroup':
            self.outgroup(options)
        elif options.subparser_name == 'propagate':
            self.propagate(options)
        elif options.subparser_name == 'fill_ranks':
            self.fill_ranks(options)
        elif options.subparser_name == 'strip':
            self.strip(options)
        elif options.subparser_name == 'rm_support':
            self.rm_support(options)
        elif options.subparser_name == 'pull':
            self.pull(options)
        elif options.subparser_name == 'append':
            self.append(options)
        elif options.subparser_name == 'pd':
            self.phylogenetic_diversity(options)
        elif options.subparser_name == 'pd_clade':
            self.phylogenetic_diversity_clade(options)
        elif options.subparser_name == 'arb_records':
            self.arb_records(options)
        else:
            self.logger.error('Unknown GenomeTreeTk command: ' + options.subparser_name + '\n')
            sys.exit(-1)

        return 0



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import math
import logging

from genometreetk.common import read_gtdb_metadata

from biolib.seq_io import read_seq
from biolib.newick import parse_label

import dendropy


class DereplicateTree(object):
    """Dereplicate tree."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger()
        
    def _derep_msa(self, msa_file, selected_taxa, output_msa):
        """Dereplicate multiple sequence alignment."""
        
        selected_taxa_labels = set()
        for taxon in selected_taxa:
            selected_taxa_labels.add(taxon.label)
        
        fout = open(output_msa, 'w')
        for seq_id, seq, annotation in read_seq(msa_file, keep_annotation=True):
            if seq_id in selected_taxa_labels:
                fout.write('>%s %s\n' % (seq_id, annotation))
                fout.write('%s\n' % seq)
        fout.close()
        
    def _derep_lineage(self, node, num_taxa_to_retain, genome_metadata):
        """Select genomes from lineage."""
        
        # rank all genomes in lineage with GTDB representatives first,
        # followed by genomes of decreasing quality
        rep_list = []
        taxa_list = []
        for leaf in node.leaf_iter():
            comp, cont, rep = genome_metadata[leaf.taxon.label]
            qual = float(comp) - 5*float(cont)
            if rep == 't' or rep == 'True' or rep == 'true':
                rep_list.append((qual, leaf.taxon))
            else:
                taxa_list.append((qual, leaf.taxon))
                
        sorted_list = sorted(rep_list, reverse=True) + sorted(taxa_list, reverse=True)
        
        taxa_to_keep = []
        for _qual, taxon in sorted_list[0:num_taxa_to_retain]:
            taxa_to_keep.append(taxon)
        
        return taxa_to_keep
        
    def _select_taxa(self, tree,
                            node_of_interest, 
                            outgroup_node, 
                            num_taxa_to_retain,
                            keep_unclassified,                            
                            genome_metadata):
        """Select genomes in named lineages on path from ingroup to outgroup."""
        
        # get most recent common ancestor of outgroup and lineage of interest
        outgroup_leaf_taxon = outgroup_node.leaf_iter().next().taxon
        lineage_of_interest_taxon = node_of_interest.leaf_iter().next().taxon
        mrca = tree.mrca(taxa=[outgroup_leaf_taxon, lineage_of_interest_taxon])
        
        # get taxon of lineage of interest
        taxa_of_interest = []
        parent = node_of_interest
        while parent != mrca:
            _support, taxon, _auxiliary_info = parse_label(parent.label)
            if taxon:
                taxa_of_interest.append(taxon)
            parent = parent.parent_node
            
        self.logger.info('Taxonomy for lineage of interest: %s' % ';'.join(taxa_of_interest))

        # select taxa from named lineages by traversing tree
        # in preorder and terminating descent at named taxa
        # not in path to lineage of interest
        selected_taxa = []
        
        stack = []
        for c in mrca.child_node_iter():
            stack.append(c)
            
        while stack:
            cur_node = stack.pop()
            
            if cur_node.is_leaf() and keep_unclassified:
                selected_taxa.append(cur_node.taxon)
            
            _support, taxon, _auxiliary_info = parse_label(cur_node.label)
            
            if taxon and taxon not in taxa_of_interest:
                # select roughly equal taxa from each child lineage to
                # enure we retain the correct depth (and the named node)
                # for this lineage
                derep_taxa = []
                num_children = sum([1 for c in cur_node.child_node_iter()])
                child_taxa_to_sample = int(math.ceil((1.0/num_children)*num_taxa_to_retain))
                for i, c in enumerate(cur_node.child_node_iter()):
                    taxa_to_sample = min(child_taxa_to_sample, num_taxa_to_retain - len(derep_taxa))
                    derep_taxa += self._derep_lineage(c, taxa_to_sample, genome_metadata)
 
                selected_taxa += derep_taxa
                self.logger.info('Selecting %d taxa from %s.' % (len(derep_taxa), taxon))
            elif cur_node == node_of_interest:
                self.logger.info('Retaining all taxa in lineage of interest.')
                for leaf in node_of_interest.leaf_iter():
                    selected_taxa.append(leaf.taxon) 
            else:
                for c in cur_node.child_node_iter():
                    stack.append(c)
                
        return selected_taxa

    def run(self, 
            input_tree, 
            lineage_of_interest, 
            outgroup,
            gtdb_metadata,
            num_taxa_to_retain,
            msa_file,
            keep_unclassified,
            output_dir):
        """Dereplicate tree.

        Parameters
        ----------
        input_tree : str
            Tree to dereplicate
        lineage_of_interest : str
            Named lineage where all taxa should be retain.
        outgroup : str
            Named lineage to use as outgroup.
        gtdb_metadata : str
            File containing metadata for taxa in tree.
        num_taxa_to_retain: int
            Taxa to retain in dereplicated lineages.
        msa_file : str
            Multiple sequence alignment to dereplicate along with tree.
        keep_unclassified : boolean
            Keep all taxa in unclassified lineages.
        output_dir:
            Output dir.
        """
        
        # read GTDB metadata
        self.logger.info('Reading metadata.')
        genome_metadata = read_gtdb_metadata(gtdb_metadata, ['checkm_completeness',
                                                              'checkm_contamination',
                                                              'gtdb_representative'])
        
        # read tree
        self.logger.info('Reading tree.')
        tree = dendropy.Tree.get_from_path(input_tree, 
                                            schema='newick', 
                                            rooting='force-rooted', 
                                            preserve_underscores=True)

        # locate node of interest and outgroup node
        self.logger.info('Identifying lineage of interest and outgroup.')
        node_of_interest = None
        outgroup_node = None
        for node in tree.preorder_node_iter():
            _support, taxon_str, _auxiliary_info = parse_label(node.label)
            
            if not taxon_str:
                continue
                
            for taxon in [t.strip() for t in taxon_str.split(';')]:
                if taxon == lineage_of_interest:
                    node_of_interest = node
                elif taxon == outgroup:
                    outgroup_node = node
                
        if not node_of_interest:
            self.logger.error('Could not find specified lineage of interest: %s' % lineage_of_interest)
            sys.exit(-1)
            
        if not outgroup_node:
            self.logger.error('Could not find outgroup: %s' % outgroup)
            sys.exit(-1)
                       
        # select taxa to retain
        self.logger.info('Selecting taxa to retain.')
        selected_taxa = self._select_taxa(tree,
                                            node_of_interest, 
                                            outgroup_node, 
                                            num_taxa_to_retain, 
                                            keep_unclassified,
                                            genome_metadata)

        self.logger.info('Retaining %d taxa.' % len(selected_taxa))
        
        # prune tree
        self.logger.info('Pruning tree.')
        tree.retain_taxa(selected_taxa)
        
        # dereplicate MSA if requested
        if msa_file:
            self.logger.info('Dereplicating MSA.')
            msa_name, msa_ext = os.path.splitext(os.path.basename(msa_file))
            output_msa = os.path.join(output_dir, msa_name + '.derep' + msa_ext)
            self._derep_msa(msa_file, selected_taxa, output_msa)
        
        # write out results
        tree_name, tree_ext = os.path.splitext(os.path.basename(input_tree))
        output_tree = os.path.join(output_dir, tree_name + '.derep' + tree_ext)
        tree.write_to_path(output_tree, 
                            schema='newick', 
                            suppress_rooting=True, 
                            unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import logging

from math import floor

import dendropy


class TreeSupport():
    """Calculate support values for clades."""

    def __init__(self):
        """Initialize."""
        self.logger = logging.getLogger()

    def subset_taxa(self, input_tree, replicate_trees, output_tree):
        """Calculate support for tree with replicates containing a subset of taxa.

        Parameters
        ----------
        input_tree : str
          Tree inferred from complete data.
        replicate_trees : iterable
          Files containing replicate trees.
        output_tree: str
          Name of output tree with support values.
        """

        tree = dendropy.Tree.get_from_path(input_tree, schema='newick', rooting='force-unrooted', preserve_underscores=True)
        for node in tree.internal_nodes():
            node.label = 0
            node.nontrivial_splits = 0

        for rep_tree_file in replicate_trees:
            print rep_tree_file
            rep_tree = dendropy.Tree.get_from_path(rep_tree_file, schema='newick', rooting='force-unrooted', preserve_underscores=True)

            rep_tree_list = dendropy.TreeList([rep_tree])

            rep_tree_taxa_set = set([x.taxon.label for x in rep_tree.leaf_nodes()])

            for i, node in enumerate(tree.internal_nodes()):
                print i
                taxa_labels = set([x.taxon.label for x in node.leaf_nodes()]).intersection(rep_tree_taxa_set)

                split = rep_tree.taxon_namespace.taxa_bitmask(labels=taxa_labels)
                normalized_split = dendropy.Bipartition.normalize_bitmask(
                                    bitmask=split,
                                    fill_bitmask=rep_tree.taxon_namespace.all_taxa_bitmask(),
                                    lowest_relevant_bit=1)

                if len(taxa_labels) > 1:
                    # tabulate results for non-trivial splits
                    node.label += int(rep_tree_list.frequency_of_bipartition(split_bitmask=normalized_split))
                    node.nontrivial_splits += 1

        for node in tree.internal_nodes():
            if node.nontrivial_splits > 0:
                node.label = str(int(floor(node.label * 100.0 / node.nontrivial_splits)))
            else:
                node.label = 'NA'

        tree.write_to_path(output_tree, schema='newick', suppress_rooting=True, unquoted_underscores=True)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import csv

"""Functions for working with NCBI genomes and metadata."""


def read_genome_dir(genome_dir_file):
    """Parse genome dir file.

    """

    genome_dirs = {}
    for line in open(genome_dir_file):
        line_split = line.rstrip().split('\t')
        genome_dirs[line_split[0]] = line_split[1]

    return genome_dirs


def read_refseq_metadata(metadata_file):
    """Parse metadata for RefSeq genomes from GTDB metadata file.

    Parameters
    ----------
    metadata_file : str
        File specifying metadata for all genomes.

    Returns
    -------
    dict : d[assembly_accession] -> taxonomy id
        Taxonomy id of assemblies.
    set
        Set of complete genomes based on NCBI assembly level.
    set
        Set of reference genomes based on NCBI RefSeq category.
    set
        Set of representative genomes based on NCBI RefSeq category.
    """

    refseq_genomes = set()
    complete_genomes = set()
    reference_genomes = set()
    representative_genomes = set()

    csv_reader = csv.reader(open(metadata_file, 'rt'))
    bHeader = True
    for row in csv_reader:
        if bHeader:
            genome_index = row.index('accession')
            assembly_level_index = row.index('ncbi_assembly_level')
            refseq_category_index = row.index('ncbi_refseq_category')
            bHeader = False
        else:
            genome_id = row[genome_index]
            if genome_id.startswith('RS_'):
                refseq_genomes.add(genome_id)

                assembly_level = row[assembly_level_index].lower()
                if assembly_level == 'complete genome':
                    complete_genomes.add(genome_id)

                refseq_category = row[refseq_category_index].lower()
                if 'reference' in refseq_category:
                    reference_genomes.add(genome_id)
                elif 'representative' in refseq_category:
                    representative_genomes.add(genome_id)

    return refseq_genomes, complete_genomes, reference_genomes, representative_genomes




###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

import os
import sys
import csv
import logging

import biolib.seq_io as seq_io
from biolib.taxonomy import Taxonomy

from genometreetk.common import (read_gtdb_metadata,
                                    read_genome_dir_file,
                                    read_gtdb_taxonomy)


class Arb(object):
    """Methods for producing files for ARB."""

    def __init__(self):
        """Initialization."""

        self.logger = logging.getLogger()
        
    def _record(self, fout, 
                    genome_id,
                    metadata_fields, 
                    metadata_values,
                    aligned_seq):
        """Write out ARB record for genome."""

        fout.write("BEGIN\n")
        fout.write("db_name=%s\n" % genome_id)
        for col_header, value in zip(metadata_fields, metadata_values):
            # replace equal signs as these are incompatible with the ARB parser
            if value:
                value = value.replace('=', '/')

            fout.write("%s=%s\n" % (col_header, value))
        
        fout.write("warning=\n")
        fout.write("aligned_seq=%s\n" % (aligned_seq))
        fout.write("END\n\n")

    def create_records(self, metadata_file, msa_file, taxonomy_file, genome_list, output_file):
        """Create ARB records from GTDB metadata."""
        
        seqs = {}
        if msa_file:
            seqs = seq_io.read(msa_file)
        
        taxonomy = {}
        if taxonomy_file:
            taxonomy = Taxonomy().read(taxonomy_file)
            
        genomes_to_keep = set()
        if genome_list:
            for line in open(genome_list):
                genomes_to_keep.add(line.strip())
                
        fout = open(output_file, 'w')
        
        delimiter = ','
        if metadata_file.endswith('.tsv'):
            delimiter = '\t'
        
        header = True
        for row in csv.reader(open(metadata_file, 'rb'), delimiter=delimiter):
            if header:
                fields = [f.lower().replace(' ', '_').replace('-', '_') for f in row[1:]]
                if taxonomy:
                    fields.append('gtdb_taxonomy')
                header = False
            else:
                genome_id = row[0]
                values = row[1:]
                if taxonomy:
                    values.append('; '.join(taxonomy[genome_id]))
                aligned_seq = seqs.get(genome_id, '')
                
                if not genomes_to_keep or genome_id in genomes_to_keep:
                    self._record(fout, genome_id, fields, values, aligned_seq)

        fout.close()
        
    def write(self, hashes, io):
        """Write data to a GreenGenes formatted files.

        Parameters
        ----------
        hashes : list of dict[feature] -> value
            GreenGenes style metadata.
        io : output stream
            Handle to output stream.
        """
        is_first = True
        for dahash in hashes:
            if is_first:
                is_first = False
            else:
                io.write('\n')

            io.write('BEGIN\n')
            for key in sorted(dahash.keys()):
                if key != 'warning' and key != 'aligned_seq':
                    io.write('='.join([key, dahash[key]]) + '\n')

            # the warning field must be the second to last as it is used
            # to indicates that the aligned sequence is to follow
            io.write('='.join(['warning', dahash.get('warning', '')]) + '\n')

            # the aligned sequence must be the last field
            io.write('='.join(['aligned_seq', dahash.get('aligned_seq', '')]) + '\n')

            io.write('END\n')
            